import re

from sqlalchemy.sql import sqltypes

from pg_database import types


SAFE_SQL_REGEX = re.compile(r"^[0-9A-Za-z_]{1,63}$")
SQL_TYPE_REGEX = re.compile(r"^[0-9A-Za-z_\(\),]+$")
SQL_USING_REGEX = re.compile(r"^[0-9A-Za-z_\(\),:']+$")


def validate_columns_in(table, column_names, empty_table, message=None):
    """
    Helper for ensuring that all column names exist in a table
    :param table: a sqlalchemy table object
    :param column_names: a list of string column names
    :param empty_table: the expected name of the table if table is None
    :param message: a custom error message (precedes list of invalid columns)
    """

    validate_table_name(table, empty_table)

    table_cols = table.columns
    error_text = message or f'Invalid column names for "{table.name}"'

    if not column_names:
        raise ValueError(f"{error_text}: empty")
    if not all(c in table_cols for c in column_names):
        invalid_columns = ", ".join(set(column_names).difference(c.name for c in table_cols))
        raise ValueError(f"{error_text}: {invalid_columns}")


def validate_column_type(column_type):
    """
    Helper for validating a column type
    :param column_type: the type to validate, which may be:
        * a string indicating the type
        * a sqlalchemy.sql.sqltypes.TypeEngine class or instance
    :see types.COLUMN_TYPE_MAP: for string values that map to types
    """

    if not column_type:
        raise ValueError(f"Invalid column type: empty")
    elif isinstance(column_type, str):
        base_type = column_type.split("(", 1)[0]
        remaining = column_type[len(base_type):]

        if base_type.lower() not in types.COLUMN_TYPE_MAP:
            raise ValueError(f"Invalid column type: {column_type}")
        if remaining and not SQL_TYPE_REGEX.match(remaining):
            raise ValueError(f"Invalid column type: {column_type}")
    else:
        is_type_class = isinstance(column_type, sqltypes.TypeEngine)
        is_type_instance = isinstance(column_type, type) and issubclass(column_type, sqltypes.TypeEngine)

        if not is_type_class and not is_type_instance:
            raise ValueError(f"Invalid column type: {column_type}")

    return column_type


def validate_pooling_params(pooling_params):
    """ Helper to supported sqlalchemy pooling options """

    supported = ("max_overflow", "pool_recycle", "pool_size", "pool_timeout")

    if not pooling_params:
        return {}
    if not isinstance(pooling_params, dict):
        raise ValueError(f"Invalid pooling params: {pooling_params}")
    if any(param not in supported for param in pooling_params):
        invalid_params = ", ".join(set(pooling_params).difference(supported))
        raise ValueError(f"Invalid pooling params: {invalid_params}")
    if any(not isinstance(pooling_params[param], int) for param in pooling_params):
        invalid_params = ", ".join(param for param in pooling_params if not isinstance(pooling_params[param], int))
        raise ValueError(f"Pooling params require integer values: {invalid_params}")

    return pooling_params


def validate_sql_params(empty_message=None, **params):
    """
    Helper for validating parameters to be passed to sqlalchemy for sql injection
    :param empty_message: a custom error message for required params
    :param params: a dict of parameter names to form the error message, for example:
        table="bad-table"                    --> "Invalid table name: bad-table"
        sql_types=["bad-type", "wrong-type"] --> "Invalid sql types: bad-type, wrong-type"
    """

    for param, param_val in params.items():

        param_text = f"{param} name" if "_" not in param else " ".join(param.split("_"))
        param_vals = param_val if isinstance(param_val, (list, tuple)) else [param_val]

        if not param_vals and empty_message:
            raise ValueError(empty_message)

        for val in param_vals:
            if not val and empty_message:
                raise ValueError(empty_message)
            if not SAFE_SQL_REGEX.match(str(val or "")):
                invalid = ",".join(str(p or "") for p in param_vals)
                raise ValueError(f"Invalid {param_text}: {invalid}")


def validate_table_name(table, empty_table=None):
    """
    Helper for validating a table object
    :param table: a sqlalchemy table object
    :param empty_table: the expected name of the table if table is None
    """

    if not getattr(table, "exists", False):
        empty_table = getattr(table, "name", empty_table)
        raise ValueError(f"No table named {empty_table}")

    return table



import datetime
import json
import logging

from frozendict import frozendict
from geoalchemy2 import types as gistypes
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import sqltypes

from .conf import settings

logger = logging.getLogger(__name__)

COLUMN_TYPE_MAP = frozendict({
    "bool": sqltypes.Boolean,
    "boolean": sqltypes.Boolean,
    "bigint": sqltypes.BigInteger,
    "binary": sqltypes.LargeBinary,
    "bytea": sqltypes.LargeBinary,
    "geometry": gistypes.Geometry,
    "geography": gistypes.Geography,
    "int": sqltypes.Integer,
    "integer": sqltypes.Integer,
    "float": sqltypes.Float,
    "date": sqltypes.Date,
    "datetime": sqltypes.DateTime,
    "decimal": sqltypes.Numeric,
    "double": sqltypes.Numeric,
    "numeric": sqltypes.Numeric,
    "number": sqltypes.Numeric,
    "json": postgresql.json.JSON,
    "jsonb": postgresql.json.JSONB,
    "raster": gistypes.Raster,
    "string": sqltypes.UnicodeText,
    "text": sqltypes.UnicodeText,
    "timestamp": sqltypes.DateTime,
    "unicode": sqltypes.UnicodeText,
    "varchar": sqltypes.UnicodeText,
})

SQL_TYPE_MAP = frozendict({
    "big_integer": "bigint",
    "binary": "bytea",
    "datetime": "timestamp",
    "double": "numeric",
    "large_binary": "bytea",
    "number": "numeric",
    "string": "text",
    "unicode": "text",
    "unicode_text": "text",
})

DATE_FORMAT = settings.date_format
DATETIME_FORMAT = settings.timestamp_format

DATE_FORMAT_MAP = frozendict({
    sqltypes.DateTime: "'{value}'::timestamp",
    sqltypes.Date: "'{value}'::date",
    datetime.datetime: DATETIME_FORMAT,
    datetime.date: DATE_FORMAT,
})


def column_type_for(column_type, default='unicode'):
    """
    Helper to map incoming column_type to a sqlalchemy.sql.sqltypes class.
    :param column_type: a string, sqltypes class or sqltypes instance to map
    :param default: the default column type to use if no mapping is possible
    """
    if isinstance(column_type, sqltypes.TypeEngine):
        return column_type
    elif isinstance(column_type, type) and issubclass(column_type, sqltypes.TypeEngine):
        return column_type
    else:
        return COLUMN_TYPE_MAP.get(column_type.lower(), COLUMN_TYPE_MAP[default])


def to_date_string(column_type, value):
    """
    Helper to convert a value of type column_type to a sqlalchemy.sql.sqltypes class.
    :param column_type: the date or datetime class, or a sqlalchemy.sql.sqltypes class
    :param value: a date or datetime instance, or a date formatted string value
    """
    str_format = DATE_FORMAT_MAP.get(type(value), DATETIME_FORMAT)
    sql_format = DATE_FORMAT_MAP[column_type]

    if isinstance(value, datetime.date):
        # It will be either a date or a datetime instance
        value = datetime.datetime.strftime(value, str_format)

    return sql_format.format(value=value)


def to_json_string(column_type, value):
    """
    Helper to convert a value of type column_type to a sqlalchemy.sql.sqltypes class.
    :param column_type: a class inheriting from sqlalchemy.sql.sqltypes.JSON
    :param value: a JSON compatible value or a string containing JSON content
    """
    if isinstance(value, dict):
        value = json.dumps(value)
    return f"'{value}'::{column_type.__name__}"


def type_to_string(column_type):
    """
    Helper to derive SQL string value from the provided type
    :param column_type: a string, or a sqlalchemy.sql.sqltypes class or instance
    """
    if hasattr(column_type, "name"):
        column_type = column_type.name
    if hasattr(column_type, "__visit_name__"):
        column_type = column_type.__visit_name__

    column_type = (column_type or "").lower()

    if column_type in SQL_TYPE_MAP:
        return SQL_TYPE_MAP[column_type]
    elif column_type in COLUMN_TYPE_MAP:
        return column_type

    logger.warning(f"type_to_string: unrecognized column type {column_type}")

    return column_type



import logging

from geoalchemy2.types import Geometry
from sqlalchemy import column, create_engine, exc, literal_column, select, table, text, Column, Table
from sqlalchemy.engine.url import URL
from sqlalchemy.schema import AddConstraint, DropConstraint, ForeignKeyConstraint, Index, MetaData
from sqlalchemy.sql import and_, func, Select

from .conf import settings
from .types import column_type_for, type_to_string
from .validation import validate_columns_in, validate_column_type, validate_pooling_params, validate_sql_params
from .validation import SQL_USING_REGEX

logger = logging.getLogger(__name__)


def get_engine(connect_args=None, pooling_args=None):
    """
    Creates a sqlalchemy engine with settings configured in Django or database config
    :param connect_args: optionally override configured settings with connection args
    :param pooling_args: optionally override configured settings with pooling args:
        max_overflow: sqlalchemy defaults to 10 connections beyond pool size
        pool_recycle: sqlalchemy defaults to no timeout (-1) in seconds
        pool_size:    sqlalchemy defaults to 5 connections
        pool_timeout: sqlalchemy defaults to 30 seconds
    :return: a sqlalchemy connection engine
    """

    connect_args = connect_args or settings.connect_args
    connect_kwargs = {"connect_args": connect_args} if connect_args else {}
    connect_kwargs.update(validate_pooling_params(pooling_args or settings.pooling_args))

    return create_engine(URL(**settings.database_info), **connect_kwargs)


def get_metadata(connect_args=None, pooling_args=None):
    """
    Instantiates sqlalchemy metadata with settings configured in Django or database config
    :param connect_args: optionally override configured settings with connection args
    :param pooling_args: see docstring on schema.get_engine
    :return: a sqlalchemy.sql.Metadata object
    """

    return _get_metadata(get_engine(connect_args, pooling_args))


def _get_metadata(meta_or_engine):
    """
    A helper to create or sync a metadata object with the database after structural changes
    :param meta_or_engine: an existing metadata to sync, or a sqlalchemy.engine to wrap
    :return: a new sqlalchemy.sql.Metadata object reflecting the current database structure
    """

    metadata = MetaData(getattr(meta_or_engine, "bind", meta_or_engine))
    metadata.reflect()
    return metadata


# Table utilities


def get_table(table_name, engine=None, ignore_missing=False):
    """
    Auto-load a table schema from the database
    :param table_name: the name of the table to load
    :param engine: an optional sqlalchemy.engine to use in the table query
    :param ignore_missing: if True, return None for missing table; otherwise raise ValueError
    :return: an existing sqlalchemy table, or raise ValueError if table doesn't exist
    """

    metadata = _get_metadata(get_engine() if engine is None else engine)

    try:
        return Table(table_name, metadata, autoload=True, autoload_with=metadata.bind)
    except exc.NoSuchTableError:
        if ignore_missing:
            logger.warning(f"get_table: no table named {table_name}")
            return None
        raise ValueError(f"No table named {table_name}")


def get_table_count(table_or_name):
    """
    Executes a count query against an existing table
    :param table_or_name: a sqlalchemy table object or the name of a table to query
    :return: the number of rows in the specified table, or raise ValueError if table doesn't exist
    """

    if isinstance(table_or_name, str):
        table = get_table(table_or_name)
    else:
        table = table_or_name

    return select([func.count()]).select_from(table).scalar()


def get_tables(table_names=None, engine=None):
    """
    Queries one or more sqlalchemy table objects from the database if they exist.
    :param table_names: optional comma-seperated column names, or a list of column names
    :param engine: an optional sqlalchemy.engine to use in the table query
    :return: all existing tables in a dict-like object, or the specified subset in a dict
    """

    all_tables = _get_metadata(get_engine() if engine is None else engine).tables

    if table_names is None:
        return all_tables
    elif isinstance(table_names, str):
        table_names = [t.strip() for t in table_names.split(",")]

    return {t: all_tables[t] for t in table_names if t in all_tables}


def table_exists(table_or_name):
    """
    Determines if a table exists in the database
    :param table_or_name: a sqlalchemy table object or the name of a table to query
    :return: True if table exists, otherwise False
    """

    if isinstance(table_or_name, str):
        table = get_table(table_or_name, ignore_missing=True)
    else:
        table = table_or_name

    return table is not None and table.exists()


def create_table(table_name, index_cols=None, dropfirst=True, engine=None, **column_types):
    """
    Creates a table in the database according to specified parameters
    :param table_name: the name of the table to create
    :param index_cols: optional comma-seperated column names, or a list or dict of column names to index
        * if index_cols is a list or string, the columns specified are indexed together:
        * if index_cols is a dict:
            - keys may be comma-separated lists of columns
            - values are valid index ops (see create_index)
            - comma-separated columns are indexed together
    :param dropfirst: drop any existing table if True, otherwise raise a ValueError if table exists
    :param engine: an optional sqlalchemy.engine to use to create the table
    :param column_types: a dict where keys represent columns and values are column types:
        * keys may have leading/trailing underscores to prevent name clashes with other function params
        * values may be strings, or a sqlalchemy.sql.sqltypes.TypeEngine class or instance
        :see types.COLUMN_TYPE_MAP: for string values that map to types
    :return: the created table
    """

    # Validate types and strip leading/trailing underscores from names to prevent arg name clashes
    column_types = {c.strip("_"): validate_column_type(t) for c, t in column_types.items()}

    validate_sql_params(table=table_name, empty_message=f"No table name specified")
    validate_sql_params(
        column_names=[cname for cname in column_types.keys()],
        empty_message=f"No column names specified for {table_name}"
    )

    meta = _get_metadata(get_engine() if engine is None else engine)
    exists = table_name in meta.tables

    if exists and not dropfirst:
        raise ValueError(f"Table already exists: {table_name}")
    elif exists and dropfirst:
        meta.tables[table_name].drop(checkfirst=True)
        meta = _get_metadata(meta)

    cols = [Column(c, column_type_for(t)) for c, t in column_types.items()]
    table = Table(table_name, meta, *cols)
    table.create()

    if index_cols:
        if isinstance(index_cols, str):
            # Basic index: single or multi-column (if comma-separated)
            create_index(table, index_cols)
        else:
            # Specific index:
            #    Single or multi-column (if any item is comma-separated)
            #    Basic index type if not a dict; otherwise use type specified by dict values
            if not isinstance(index_cols, dict):
                index_cols = {}.fromkeys(index_cols)
            for col_names, index_op in index_cols.items():
                create_index(table, col_names, index_op=index_op)

    return table


def drop_table(table_or_name, checkfirst=True):
    """
    Drops a table if it exists in the database; if not a warning is logged
    :param table_or_name: a sqlalchemy table object or the name of a table to drop
    :param checkfirst: if True, log warnings for missing table; otherwise raise ValueError
    """

    if isinstance(table_or_name, str):
        table = get_table(table_or_name, ignore_missing=checkfirst)
    else:
        table = table_or_name

    exists = table is not None and table.exists()

    if not checkfirst and not exists:
        raise ValueError(f"No table named {table_or_name}")
    elif not exists:
        logger.warning(f"drop_table: no table found named {table_or_name}")
    else:
        logger.info(f"drop_table: dropping table named {table.name}")
        table.drop(checkfirst=checkfirst)


# Column utilities


def alter_column_type(table_or_name, column_name, new_type, using=None):
    """
    Alter a column existing in a given table
    :param table_or_name: a sqlalchemy table object or the name of a table with a column to alter
    :param column_name: the name of the column in the table to alter
    :param new_type: indicates what type the column should be updated to, which may be:
        * a string indicating the type
        * a sqlalchemy.sql.sqltypes.TypeEngine class or instance
        :see types.COLUMN_TYPE_MAP: for string values that map to types
    :param using: an optional, custom SQL expression to follow double colon for column data conversion:
        ALTER TABLE t
        ALTER COLUMN c TYPE geometry     // new_type="geometry"
        USING c::geometry(Polygon:4326)  // using="geometry(Polygon:4326)"
    """

    if isinstance(table_or_name, str):
        table_name = table_or_name
        sql_engine = get_engine()
    else:
        table_name = table_or_name.name
        sql_engine = table_or_name.bind

    validate_sql_params(table=table_name, column=column_name)
    validate_column_type(column_type=new_type)

    if using and not SQL_USING_REGEX.match(using):
        raise ValueError(f"Invalid column conversion: {using}")

    new_type = type_to_string(new_type)

    if using is None:
        old_type = type_to_string(_get_metadata(sql_engine).tables[table_name].columns[column_name].type)

        if new_type == "bytea" and old_type in ("json", "jsonb"):
            using = f"{column_name}::text::{new_type}"
        elif new_type in ("bool", "boolean"):
            using = f"CASE WHEN {column_name}::int=0 THEN FALSE WHEN {column_name} IS NULL THEN NULL ELSE TRUE END"
        elif new_type in ("json", "jsonb"):
            from_bytea = f"convert_from({column_name},'UTF8')::{new_type}"
            from_other = f"{column_name}::text::{new_type}"
            using = from_bytea if old_type == "bytea" else from_other
        else:
            using = f"{column_name}::{new_type}"

    alter_sql = f"ALTER TABLE {table_name} "
    alter_sql += f"ALTER COLUMN {column_name} TYPE {new_type} USING {using}"

    logger.info(f"alter_column_type: altering {table_name}.{column_name} to {new_type}")

    with sql_engine.connect() as conn:
        conn.execute(text(alter_sql).execution_options(autocommit=True))


def create_column(table_or_name, column_name, column_type, checkfirst=False, default=None, nullable=False):
    """
    Alter a column existing in a given table
    :param table_or_name: a sqlalchemy table object or the name of a table on which to create a column
    :param column_name: the name of the column to create
    :param column_type: indicates what type the column should be:
        * may be a string indicating the type
        * may also be a sqlalchemy.sql.sqltypes.TypeEngine class or instance
        :see types.COLUMN_TYPE_MAP: for string values that map to types
    :param checkfirst: if True, log warning if column exists; otherwise raise a sqlalchemy error
    :param default: a default value to assign to the column
    :param nullable: make the column support NULL if True, otherwise NOT NULL
    """

    if isinstance(table_or_name, str):
        table_name = table_or_name
        sql_engine = get_engine()
    else:
        table_name = table_or_name.name
        sql_engine = table_or_name.bind

    validate_sql_params(table=table_name, column=column_name)
    validate_column_type(column_type=column_type)

    str_default = isinstance(default, str)
    if str_default:
        validate_sql_params(default_value=default)

    alter_table = f"ALTER TABLE {table_name} "
    column_type = type_to_string(column_type)

    if checkfirst:
        add_column = f"ADD COLUMN IF NOT EXISTS {column_name} {column_type}"
    else:
        add_column = f"ADD COLUMN {column_name} {column_type}"

    constraints = ["NULL" if nullable else "NOT NULL"]

    if str_default:
        constraints.append(f"DEFAULT '{default}'")
    elif default is not None:
        constraints.append(f"DEFAULT {default}")

    create_sql = "{alter_table} {add_column} {constraints}".format(
        alter_table=alter_table, add_column=add_column, constraints=" ".join(constraints)
    )
    logger.info(f"create_column: creating {table_name}.{column_name} as {column_type}")

    with sql_engine.connect() as conn:
        conn.execute(text(create_sql).execution_options(autocommit=True))


def create_tsvector_column(table_or_name, column_name, column_names, index_name=None):
    """
    Creates a single text tsvector column from a list of columns
    NOTE: compatible with PostgreSQL version 12.* and above
    TODO: update with sqlalchemy implementation when available
    :see: https://github.com/sqlalchemy/sqlalchemy/pull/4896
    """

    if isinstance(table_or_name, str):
        table_name = table_or_name
        sql_engine = get_engine()
    else:
        table_name = table_or_name.name
        sql_engine = table_or_name.bind

    if isinstance(column_names, str):
        column_names = [c.strip() for c in column_names.split(",")]

    validate_sql_params(table=table_name, column=column_name, column_names=column_names)

    concatenated = "||' '||".join(column_names)
    generate_sql = "\n".join(
        (
            f"ALTER TABLE {table_name}",
            f"  ADD COLUMN {column_name} tsvector",
            "    GENERATED ALWAYS AS (",
            f"      TO_TSVECTOR('english', {concatenated})",
            "    ) STORED;",
        )
    )
    logger.info(f"create_tsvector_column: executing generate column query:\n{generate_sql}")

    with sql_engine.connect() as conn:
        conn.execute(generate_sql)

    create_index(table_or_name, column_names, index_name, "to_tsvector")


def drop_column(table_or_name, column_name, checkfirst=False):
    """
    Execute SQL to drop a column
    :param table_or_name: a sqlalchemy table object or the name of a table on which to create a column
    :param column_name: the name of the column to drop
    :param checkfirst: check if column exists if True, otherwise raise a sqlalchemy error if it doesn't
    """

    if isinstance(table_or_name, str):
        table_name = table_or_name
        sql_engine = get_engine()
    else:
        table_name = table_or_name.name
        sql_engine = table_or_name.bind

    validate_sql_params(table=table_name, column=column_name)

    if checkfirst:
        drop_sql = f"ALTER TABLE {table_name} DROP COLUMN IF EXISTS {column_name}"
    else:
        drop_sql = f"ALTER TABLE {table_name} DROP COLUMN {column_name}"

    logger.info(f"drop_column: dropping {table_name}.{column_name}")

    with sql_engine.connect() as conn:
        conn.execute(text(drop_sql).execution_options(autocommit=True))


# Index utilities


def create_index(table_or_name, column_names, index_name=None, index_op=None):
    """
    Creates a database index on the specified table for one or more column names

    :param table_or_name: a sqlalchemy table object or the name of a table
    :param column_names: one or more comma-seperated column names, or a list of column names
    :param index_name: optionally override the default index name generated from column_names
        It may be necessary to override the index name to ensure it is under 63 characters;
        otherwise, index creation will fail with an error from PostgreSQL.
        A conventional index name will follow this pattern: "{table_name}_{column_names}_idx"
    :param index_op: any SQL function available in Postgresql (using sqlalchemy.sql.func.*):
        * supported index types are:
            - coalesce:    CREATE INDEX <idx> ON <table> COALESCE(<col1>,<col2>,'');
            - spatial  :   CREATE INDEX <idx> ON <table> USING GIST (<col>);
            - json_full:   CREATE INDEX <idx> ON <table> USING GIN (<col>);
            - json_path:   CREATE INDEX <idx> ON <table> USING GIN (<col> jsonb_path_ops);
            - to_tsvector: CREATE INDEX <idx> ON <table> USING GIN(to_tsvector('english', <col1>||' '||<col2>));
            - unique:      CREATE UNIQUE INDEX <idx> ON <table> (<col1>,<col2>);
    """

    if isinstance(table_or_name, str):
        table = get_table(table_or_name)
    else:
        table = table_or_name

    if isinstance(column_names, str):
        column_names = [c.strip() for c in column_names.split(",")]

    validate_columns_in(table, column_names, empty_table=table_or_name, message="Invalid index column names")

    if len(column_names) > 1 and (index_op == "spatial" or (index_op or "").startswith("json")):
        raise ValueError(f"Invalid index operation for multiple columns: {index_op}")
    elif index_op == "spatial" and not isinstance(table.columns[column_names[0]].type, Geometry):
        column_type = str(table.columns[column_names[0]].type)
        raise ValueError(f"Invalid column type for spatial index: {column_type}")

    if not index_name:
        column_index = "_".join(col for col in column_names)
        if not index_op:
            index_name = f"{table.name}_{column_index}_idx"
        else:
            index_name = f"{table.name}_{column_index}_{index_op}_idx"

    index_kwargs = {"_table": table}

    if not index_op:
        expressions = [col for col in column_names if col in table.columns]
    elif index_op == "coalesce":
        # Coalesce takes an indefinite list of column names and a default value
        expressions = (func.coalesce(*(column_names + [""])),)
    elif index_op == "spatial":
        # Generates CREATE INDEX <idx> ON <table> USING GIST (<col>);
        index_kwargs["postgresql_using"] = "GIST"
        expressions = (table.columns.get(column_names[0]),)
    elif index_op == "json_full":
        # Generates CREATE INDEX <idx> ON <table> USING GIN (<col>);
        index_kwargs["postgresql_using"] = "GIN"
        expressions = (table.columns.get(column_names[0]),)
    elif index_op == "json_path":
        # Generates CREATE INDEX <idx> ON <table> USING GIN (<col> jsonb_path_ops);
        index_kwargs["postgresql_using"] = "GIN"
        expressions = (text("{col} jsonb_path_ops".format(col=column_names[0])),)
    elif index_op == "to_tsvector":
        index_kwargs["postgresql_using"] = "GIN"
        expressions = (func.to_tsvector("english", text("||' '||".join(column_names))),)
    elif index_op == "unique":
        index_kwargs["unique"] = True
        expressions = [col for col in column_names if col in table.columns]
    else:
        raise ValueError(f"Unsupported index type: {index_op}")

    logger.info(f"create_index: creating index {index_name} on {table.name}")
    Index(index_name, *expressions, **index_kwargs).create()


def drop_index(table_or_name, index_name=None, column_names=None, ignore_errors=True):
    """
    Drops a database index by name for a specified table if it exists

    :param table_or_name: a sqlalchemy table object or the name of a table
    :param index_name: EITHER specify an index name to drop
    :param column_names: OR derive index name from a list or string of comma-seperated column names
    :param ignore_errors: if True, will log warning for database errors; otherwise will raise them

    If both index_name and column_names are provided, index_name takes precedence
    """

    if isinstance(table_or_name, str):
        table = get_table(table_or_name, ignore_missing=ignore_errors)
    else:
        table = table_or_name

    if table is None:
        error = f"No table named {table_or_name}"
    elif column_names is None and index_name is None:
        error = "No index name provided"
    else:
        error = None

    if error:
        if not ignore_errors:
            raise ValueError(error)
        logger.warning(f"drop_index: {error}".lower())
        return

    if index_name is not None:
        column_index = None
    elif isinstance(column_names, str):
        column_index = "_".join(col.strip() for col in column_names.split(","))
    else:
        column_index = "_".join(column_names)

    if column_index is not None:
        index_name = f"{table.name}_{column_index}_idx"

    logger.info(f"drop_index: dropping index named {index_name}")

    try:
        Index(index_name, _table=table).drop()
    except exc.ProgrammingError:
        if not ignore_errors:
            raise
        logger.warning(f"drop_index: no index found named {index_name}")


def has_index(table_or_name, index_name):
    """
    Queries `pg_indexes` to see if an index exists, expression-based or otherwise, for a table.

    :param table_or_name: a sqlalchemy table object or the name of a table
    :param index_name: the name of the index to check for
    """

    if isinstance(table_or_name, str):
        table_name = table_or_name
        sql_engine = get_engine()
    else:
        table_name = table_or_name.name
        sql_engine = table_or_name.bind

    index_query = (
        Select([literal_column("1")], distinct=True)
        .select_from(table("pg_indexes"))
        .where(and_(column("tablename") == table_name, column("indexname") == index_name))
    )

    with sql_engine.connect() as conn:
        return bool(conn.execute(index_query).first())


def create_foreign_key(table_or_name, column_name, related_column_or_name):
    """
    Creates a foreign key on the specified table given two column names

    :param table_or_name: a sqlalchemy table object or the name of a table referencing another one
    :param column_name: the name of the column in the referencing table
    :param related_column_or_name: the name of the column in a referenced table
        * this value must be a string including the table name: 'other_table.col_to_reference'
    """

    if isinstance(table_or_name, str):
        table = get_table(table_or_name)
    else:
        table = table_or_name

    if not isinstance(related_column_or_name, str):
        related = related_column_or_name
    else:
        rel, col = related_column_or_name.split(".")
        related = get_tables(engine=table.bind)
        related = getattr(related.get(rel), "columns", {}).get(col)

    validate_columns_in(table, [column_name], empty_table=table_or_name)

    if related is None:
        raise ValueError(f"No related column named {related_column_or_name}")
    else:
        col_name = f"{table.name}.{column_name}"
        rel_name = f"{related.table.name}.{related.name}"
        logger.info(f"create_foreign_key: adding FK to {col_name} on {rel_name}")

        ddl = AddConstraint(ForeignKeyConstraint([table.columns[column_name]], [related]))
        ddl.execute(table.bind, ddl.target)


def drop_foreign_key(table_or_name, fk_or_name, checkfirst=False):
    """
    Drops a foreign key for a specified table if it exists
    :param table_or_name: a sqlalchemy table object or the name of a table
    :param fk_or_name: a sqlalchemy foreign key constraint, or the name of one
    :param checkfirst: if True, log warnings for missing table or FK; otherwise raise ValueError
    """

    if isinstance(table_or_name, str):
        table = get_table(table_or_name, ignore_missing=checkfirst)
    else:
        table = table_or_name

    if table is None:
        if checkfirst:
            logger.info(f"drop_foreign_key: no table named {table_or_name}")
            return
        raise ValueError(f"No table named {table_or_name}")

    if not isinstance(fk_or_name, str):
        fk = getattr(fk_or_name, "constraint", fk_or_name) or None
    else:
        fks = {fk.name.lower(): fk for fk in table.foreign_key_constraints}
        fk = fks.get(fk_or_name.lower())

    if fk is None:
        if checkfirst:
            logger.info(f"drop_foreign_key: no such foreign key in table {table.name}")
            return
        raise ValueError(f"No such foreign key in table {table.name}")

    logger.info(f"drop_foreign_key: dropping FK to {fk.name} from table {table.name}")

    ddl = DropConstraint(fk)
    ddl.execute(table.bind, ddl.target)



import secrets

DEBUG = True
SECRET_KEY = secrets.token_urlsafe(50)

DATABASES = {
    "default": {
        # Missing database and user name
        "ENGINE": "django.db.backends.postgresql",
        "USER": "",
    }
}



import json
import secrets

from pathlib import Path


BASE_DIR = Path(__file__).resolve().parent

CONFIG = {}
config_file = Path(BASE_DIR / "test_config.json").resolve()
if config_file and config_file.is_file():
    with open(config_file) as f:
        CONFIG = json.loads(f.read())

DEBUG = True
SECRET_KEY = secrets.token_urlsafe(50)

DATABASES = {
    "nondefault": {
        "ENGINE": "django.contrib.gis.db.backends.postgis",
        "NAME": CONFIG.get("database-name"),
        "USER": "django",
        "OPTIONS": {"sslmode": "allow"},
    },
}

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.gis',
]



import json
import logging
import os
import pathlib

from frozendict import frozendict

logger = logging.getLogger(__name__)

DJANGO_SETTINGS_VAR = "DJANGO_SETTINGS_MODULE"
ENVIRONMENT_VARIABLE = "DATABASE_CONFIG_JSON"

DEFAULT_DJANGO_DB = "unspecified"
DEFAULT_ENGINE = "postgresql"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 5432
DEFAULT_USER = "postgres"

DEFAULT_DATE_FORMAT = "%Y-%m-%d"
DEFAULT_TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S"

DATABASE_PROPS = frozenset({
    # Supported database info properties
    "database_engine", "engine", "drivername",
    "database_name", "name", "database",
    "database_port", "port",
    "database_host", "host",
    "database_user", "user", "username",
    "database_password", "password",
    # Supported database query properties
    "connect_args", "pooling_args",
    # Supported django database properties
    "ENGINE", "django_engine",
    "NAME", "django_name",
    "PORT", "django_port",
    "HOST", "django_host",
    "USER", "django_user",
    "PASSWORD", "django_password",
    "OPTIONS", "django_options",
})
SUPPORTED_CONFIG = frozendict({
    # Required database configuration
    "database-name": None,
    # Optional database configuration
    "database-engine": DEFAULT_ENGINE,
    "database-host": DEFAULT_HOST,
    "database-port": DEFAULT_PORT,
    "database-user": "postgres",
    "database-password": None,
    # Optionally specify configured Django database
    "django-db-key": DEFAULT_DJANGO_DB,
    # Other non-required options:
    "connect-args": None,
    "pooling-args": None,
    "date-format": DEFAULT_DATE_FORMAT,
    "timestamp-format": DEFAULT_TIMESTAMP_FORMAT,
})


try:
    from django.utils.functional import empty
    EMPTY = empty
    DJANGO_INSTALLED = True
except ImportError:
    EMPTY = object()
    DJANGO_INSTALLED = False


class PgDatabaseSettings(object):
    """
    A class to consolidate settings derived from the following sources:
        * A JSON file identified by the DATABASE_CONFIG_JSON environment variable
        * An appropriately configured Django settings file if Django is installed

    Precedence of settings duplicated in both sources is as follows:
        * If no DATABASE_CONFIG_JSON file is specified, use Django.DATABASES
        * If DATABASE_CONFIG_JSON specifies a django-db-key, then use that Django database
        * Otherwise use the keys defined in DATABASE_CONFIG_JSON

    This is how database connection info is mapped between sources:
        | config key        | django   | database_info | settings property | default values |
        |:------------------|:---------|:--------------|:------------------|:---------------|
        | database-engine   | ENGINE   | drivername    | database_engine   | "postgresql"   |
        | database-host     | HOST     | host          | database_host     | "127.0.0.1"    |
        | database-port     | PORT     | port          | database_port     | 5432           |
        | database-name     | NAME     | database      | database_name     | <required>     |
        | database-user     | USER     | username      | database_user     | "postgres"     |
        | database-password | PASSWORD | password      | database_password | None           |

    This is how other database options are mapped:
        | config key       | django   | settings property | default values        |
        |:-----------------|:---------|:------------------|:----------------------|
        | connect-args     | OPTIONS  | connect_args      | None                  |
        | pooling-args     | None     | pooling_args      | <see sqlalchemy docs> |
        | django-db-key    | None     | django_db_key     | "default"             |
        | date-format      | None     | date_format       | "%Y-%m-%d"            |
        | timestamp-format | None     | timestamp_format  | "%Y-%m-%d %H:%M:%S"   |

    Of the above, only database-name and database-user are required.
    The others either have defaults or are not required for database connection.
    """

    empty = EMPTY

    _database_info = None
    _database_config = None
    _django_settings = None
    _django_database = None

    def __init__(self):
        self._init_database_config()
        self._init_django_settings()

    def _init_database_config(self):
        """ Initialize a valid database configuration if one has been provided """

        if self._database_config is not None:
            return

        self._database_config = frozendict()

        if ENVIRONMENT_VARIABLE not in os.environ:
            logger.debug("Database configuration not provided")

        elif not (os.environ[ENVIRONMENT_VARIABLE] or "").endswith(".json"):
            config_path = os.environ[ENVIRONMENT_VARIABLE]
            raise EnvironmentError(f"Invalid database configuration file: {config_path}")

        else:
            try:
                config_path = os.environ[ENVIRONMENT_VARIABLE]
                config_file = pathlib.Path(config_path).expanduser().resolve(strict=True)
                with open(config_file) as config_json:
                    database_config = json.loads(config_json.read())

            except FileNotFoundError:
                raise EnvironmentError(f"Database configuration file does not exist: {config_path}")
            except ValueError:
                raise EnvironmentError(f"Database configuration file does not contain JSON: {config_path}")

            # Remove values from overridden configuration if not present in supported configuration
            database_config = {k: v for k, v in database_config.items() if k in SUPPORTED_CONFIG}
            # Assign defaults in supported configuration if not present in overridden configuration
            database_config.update({k: v for k, v in SUPPORTED_CONFIG.items() if k not in database_config})

            self._database_config = frozendict(database_config)

    def _init_django_settings(self):
        """ Capture properly configured Django settings if Django is present and configured """

        if self._django_settings is not None:
            return

        self._django_settings = EMPTY

        if DJANGO_INSTALLED:
            try:
                from django.core.exceptions import ImproperlyConfigured
                from django.conf import settings as _django_settings

                # Raises ImproperlyConfigured without DJANGO_SETTINGS_MODULE
                getattr(_django_settings, 'DATABASES', None)
                self._django_settings = _django_settings

            except ImproperlyConfigured as ex:
                logger.debug(f"Django settings are not configured: {ex}")

    def _init_database_info(self):
        """
        Lazily initializes database info from configuration file or Django settings.
        The implementation is lazy in that database info isn't determined at instantiation time.
        """

        if self._database_info is not None:
            return

        self._database_info = frozendict()
        self._django_database = frozendict()

        # Validate and apply database configuration

        # Reference DATABASES at run time to support django.test.override_settings
        django_databases = getattr(self._django_settings, "DATABASES", None)

        if not self._database_config and not django_databases:
            raise EnvironmentError(
                "No database configuration available:\n"
                f'Hint: have you set the "{ENVIRONMENT_VARIABLE}" environment variable?'
            )
        elif not django_databases:
            logger.debug("Using provided database configuration")

        elif not self._database_config or self._database_config.get("django-db-key") != DEFAULT_DJANGO_DB:
            logger.debug("Applying Django database configuration")

            django_db_key = self._database_config.get("django-db-key") or "default"
            if django_db_key not in django_databases:
                raise EnvironmentError(f'No Django database configured for: "{django_db_key}"')

            django_database = {}.fromkeys(("ENGINE", "HOST", "PORT", "NAME", "USER", "PASSWORD"))
            django_database.update(django_databases[django_db_key])

            # Update configuration with values from Django config or defaults
            database_config = dict(self._database_config)
            database_config.update({
                "database-engine": DEFAULT_ENGINE,
                "database-host": django_database.get("HOST") or DEFAULT_HOST,
                "database-port": django_database.get("PORT") or DEFAULT_PORT,
                "database-name": django_database.get("NAME"),
                "database-user": django_database.get("USER"),
                "database-password": django_database.get("PASSWORD"),
                "connect-args": django_database.get("OPTIONS", database_config.get("connect-args"))
            })

            self._database_config = frozendict(database_config)
            self._django_database = frozendict({f"django_{k}".lower(): v for k, v in django_database.items()})

        for database_key in ("database-name", "database-user"):
            if self._database_config[database_key] is None:
                raise EnvironmentError(f'Database configuration missing required key: "{database_key}"')

        database_info = {
            # Properties named after sqlalchemy.engine.url.URL params
            "database": self._database_config["database-name"],
            "drivername": self._database_config["database-engine"],
            "host": self._database_config["database-host"],
            "port": self._database_config["database-port"],
            "username": self._database_config["database-user"],
            "password": self._database_config["database-password"],
        }

        if database_info["password"] is None:
            database_info["host"] = None

        self._database_info = frozendict(database_info)

    @property
    def database_info(self):
        if self._database_info is None:
            self._init_database_info()
        return self._database_info

    @property
    def django_database(self):
        if self._django_database is None:
            self._init_database_info()
        return self._django_database

    @property
    def database_engine(self):
        """ Routes "database_engine" to database info property: drivername """
        return (self._database_info or self.database_info)["drivername"]

    @property
    def database_name(self):
        """ Routes "database_name" to database info property: database """
        return (self._database_info or self.database_info)["database"]

    @property
    def database_user(self):
        """ Routes "database_user" to database info property: username """
        return (self._database_info or self.database_info)["username"]

    def __getattr__(self, name):
        """ Exposes database info and config consistently with precedence  """

        if self._database_info is None and name in DATABASE_PROPS:
            self._init_database_info()

        # First check database info initialized from config or django

        if self._database_info:
            if name in self._database_info:
                return (self._database_info or self.database_info)[name]

            if name in ("engine", "name", "user"):
                return getattr(self, f"database_{name}")

            if name.startswith("database_"):
                prop = name[len("database_"):]
                if prop in self._database_info:
                    return (self._database_info or self.database_info)[prop]

        # Next check for references to django database (i.e. ENGINE, django_engine)

        if self._django_database:
            if name in self._django_database:
                return (self._django_database or self.django_database)[name]

            if name.isupper():
                prop = f"django_{name.lower()}"
                if prop in self._django_database:
                    return (self._django_database or self.django_database)[prop]

        # Finally check for any other configured properties (i.e. connect_args, date_format, Etc.)

        if "-" not in name:
            prop = name.replace("_", "-")
            if prop in self._database_config:
                return self._database_config[prop]

        return None


settings = PgDatabaseSettings()



import copy
import datetime
import decimal
import json
import os
import pytest
import random
import warnings

from geoalchemy2 import types as gistypes
from sqlalchemy import create_engine, Column, Index, Table, select
from sqlalchemy.dialects import postgresql
from sqlalchemy.engine.url import URL
from sqlalchemy.exc import ArgumentError, SAWarning
from sqlalchemy.sql import func, text, sqltypes
from sqlalchemy.schema import AddConstraint, DropConstraint, ForeignKeyConstraint, MetaData

from pg_database import conf
from pg_database import schema
from pg_database import sql
from pg_database import types
from pg_database import validation


DJANGO_SETTINGS_VAR = conf.DJANGO_SETTINGS_VAR
ENVIRONMENT_VARIABLE = conf.ENVIRONMENT_VARIABLE
DEFAULT_ENGINE = conf.DEFAULT_ENGINE
DEFAULT_PORT = conf.DEFAULT_PORT
DEFAULT_USER = conf.DEFAULT_USER

DEFAULT_DJANGO_DB = conf.DEFAULT_DJANGO_DB
DEFAULT_DATE_FORMAT = conf.DEFAULT_DATE_FORMAT
DEFAULT_TIMESTAMP_FORMAT = conf.DEFAULT_TIMESTAMP_FORMAT
POSTGIS_TABLE_NAME = "spatial_ref_sys"

DATABASE_INFO = conf.settings.database_info
DATABASE_NAME = conf.settings.database_name
SITE_TABLE_NAME = "site_table"
ALL_TEST_TABLES = (SITE_TABLE_NAME, f"test_{SITE_TABLE_NAME}", f"tmp_{SITE_TABLE_NAME}")

GEOM_COL_TYPES = {
    "test_geom": ("geometry", "geometry(POINT,4326)"),
    "test_line": ("geometry", "geometry(LINESTRING,4326)"),
    "test_poly": ("geometry", "geometry(POLYGON,4326)"),
}
SITE_COL_TYPES = {
    "pk": sqltypes.Integer,
    "obj_order": sqltypes.Numeric,
    "obj_hash": sqltypes.Text,
    "site_json": postgresql.json.JSONB,
    "test_bool": sqltypes.Boolean,
    "test_date": sqltypes.Date,
    "test_json": postgresql.json.JSONB,
    "test_none": sqltypes.Text,
}
SITE_TABLE_COLS = (
    "pk",
    "obj_hash",
    "obj_order",
    "site_addr",
    "site_city",
    "site_state",
    "site_zip",
    "site_json",
    "countyname",
    "test_bool",
    "test_date",
    "test_json",
    "test_none",
    "test_geom",
    "test_line",
    "test_poly",
)
SEARCHABLE_COLS = ("obj_hash", "site_addr", "site_city", "site_state", "site_zip", "countyname")

SITE_TEST_DATA = [
    {
        "pk": "0",
        "obj_hash": "0-0-7350-12224",
        "obj_order": 0,
        "site_addr": "7350 STATE ST",
        "site_city": "ALBANY",
        "site_state": "NY",
        "site_zip": "12224",
        "site_json": {"number": "7350", "street": "STATE ST", "city": "ALBANY", "state": "NY"},
        "countyname": "Albany",
        "test_bool": 1,
        "test_date": datetime.datetime.strptime("2020-02-02 00:00:00", types.DATETIME_FORMAT),
        "test_json": {"data": [1, 2, 3]},
        "test_none": "null",
        "test_geom": None,
        "test_line": None,
        "test_poly": None,
    },
    {
        "pk": "1",
        "obj_hash": "1-9-7350-12224",
        "obj_order": 9,
        "site_addr": "7350 STATE ST",
        "site_city": "ALBANY",
        "site_state": "NY",
        "site_zip": "12224",
        "site_json": {"number": "7350", "street": "STATE ST", "city": "ALBANY", "state": "NY"},
        "countyname": "Albany",
        "test_bool": 0,
        "test_date": datetime.datetime.strptime("2010-01-01", types.DATE_FORMAT),
        "test_json": ["one", "two", "three"],
        "test_none": None,
        "test_geom": None,
        "test_line": None,
        "test_poly": None,
    },
    {
        "pk": "2",
        "obj_hash": "2-1-7350-12224",
        "obj_order": 1,
        "site_addr": "7350 STATE ST APT 1100",
        "site_city": "ALBANY",
        "site_state": "NY",
        "site_zip": "12224",
        "site_json": {"number": "7350", "street": "STATE ST", "unit": "1100", "city": "ALBANY", "state": "NY"},
        "countyname": "Albany",
        "test_bool": 0.0,
        "test_date": datetime.datetime.strptime("2020-01-02 00:00:00", types.DATETIME_FORMAT).date(),
        "test_json": [],
        "test_none": "",
        "test_geom": None,
        "test_line": None,
        "test_poly": None,
    },
    {
        "pk": "3",
        "obj_hash": "3-8-7450-12224",
        "obj_order": 8,
        "site_addr": "7450 STATE ST",
        "site_city": "ALBANY",
        "site_state": "NY",
        "site_zip": "12224",
        "site_json": {"number": "7450", "street": "STATE ST", "city": "ALBANY", "state": "NY"},
        "countyname": "Albany",
        "test_bool": 1.0,
        "test_date": datetime.datetime.strptime("2010-02-02", types.DATE_FORMAT).date(),
        "test_json": {},
        "test_none": "[]",
        "test_geom": None,
        "test_line": None,
        "test_poly": None,
    },
    {
        "pk": "4",
        "obj_hash": "4-2-7450-12224",
        "obj_order": 2,
        "site_addr": "7450 STATE ST APT 1101",
        "site_city": "ALBANY",
        "site_state": "NY",
        "site_zip": "12224",
        "site_json": {"number": "7450", "street": "STATE ST", "unit": "1101", "city": "ALBANY", "state": "NY"},
        "countyname": "Albany",
        "test_bool": False,
        "test_date": "2020-02-01",
        "test_json": {"id": 4, "hash": "4-2-7450-12224"},
        "test_none": "{}",
        "test_geom": None,
        "test_line": None,
        "test_poly": None,
    },
    {
        "pk": "5",
        "obj_hash": "5-7-7550-12224",
        "obj_order": 7,
        "site_addr": "7550 STATE AVE",
        "site_city": "ALBANY",
        "site_state": "NY",
        "site_zip": "12224",
        "site_json": {"number": "7550", "street": "STATE ST", "city": "ALBANY", "state": "NY"},
        "countyname": "Albany",
        "test_bool": None,
        "test_date": "2010-02-01",
        "test_json": {"county": "Albany", "city": "ALBANY"},
        "test_none": "0",
        "test_geom": None,
        "test_line": None,
        "test_poly": None,
    },
    {
        "pk": "6",
        "obj_hash": "6-3-7550-12224",
        "obj_order": 3,
        "site_addr": "7550 STATE ST NORTH",
        "site_city": "ALBANY",
        "site_state": "NY",
        "site_zip": "12224",
        "site_json": {"number": "7550", "street": "STATE ST NORTH", "city": "ALBANY", "state": "NY"},
        "countyname": "Albany",
        "test_bool": True,
        "test_date": "2020-04-04",
        "test_json": {"state": "NY", "zip": "https://calendar.google.com/"},
        "test_none": "NaN",
        "test_geom": None,
        "test_line": None,
        "test_poly": None,
    },
    {
        "pk": "7",
        "obj_hash": "7-6-7550-12224",
        "obj_order": 6,
        "site_addr": "7550 STATE ST APT 1",
        "site_city": "ALBANY",
        "site_state": "NY",
        "site_zip": "12224",
        "site_json": {"number": "7550", "street": "STATE ST", "unit": "1", "city": "ALBANY", "state": "NY"},
        "countyname": "Albany",
        "test_bool": 1,
        "test_date": "2020-03-03",
        "test_json": "7-6-7550-12224",
        "test_none": "false",
        "test_geom": None,
        "test_line": None,
        "test_poly": None,
    },
    {
        "pk": "8",
        "obj_hash": "8-4-7550-12224",
        "obj_order": 4,
        "site_addr": "7550 STATE ST STE CML5",
        "site_city": "ALBANY",
        "site_state": "NY",
        "site_zip": "12224",
        "site_json": {"number": "7550", "street": "STATE ST STE CML5", "city": "ALBANY", "state": "NY"},
        "countyname": "Albany",
        "test_bool": True,
        "test_date": "2012-12-12",
        "test_json": 8,
        "test_none": "undefined",
        "test_geom": None,
        "test_line": None,
        "test_poly": None,
    },
]
SITE_DATA_DICT = {obj["pk"]: obj for obj in SITE_TEST_DATA}


def assert_index(meta, table_name, column_name=None, index_name=None, exists=True):

    assert index_name or column_name

    test_table = refresh_metadata(meta).tables[table_name]

    if column_name:
        assert exists == any([column_name in i.columns for i in test_table.indexes])
    if index_name:
        index_query = (
            "SELECT TRUE FROM pg_catalog.pg_indexes "
            f"WHERE tablename = '{table_name}' AND indexname = '{index_name}'"
        )
        assert exists == (meta.bind.execute(index_query).scalar() or False)


def assert_records(meta, table_name, target_data, target_columns=None, target_count=None):

    def to_date(val):
        if type(val) is datetime.date:
            return val
        elif isinstance(val, datetime.datetime):
            return val.date()
        elif isinstance(val, str):
            return datetime.datetime.strptime(val.split()[0], types.DATE_FORMAT).date()

    test_table = refresh_metadata(meta).tables[table_name]

    # Test the expected count of records
    target_count = len(target_data) if target_count is None else target_count
    assert select([func.count()]).select_from(test_table).scalar() == target_count

    # Test the value of each record as identified by "pk"
    for record in (dict(row) for row in test_table.select().execute()):
        pk = record["pk"]
        target = target_data[pk if pk in target_data else str(pk)]
        fields = target_columns or list(target)

        for field in (f for f in record if f in fields):
            target_val = target[field]
            record_val = record[field]

            if isinstance(record_val, bool):
                assert record_val == bool(target_val)
            elif isinstance(record_val, datetime.date):
                assert to_date(record_val) == to_date(target_val)
            elif isinstance(record_val, (int, float, decimal.Decimal)):
                assert float(record_val) == float(target_val)
            elif isinstance(record_val, (dict, list, set, tuple)):
                assert record_val == target_val
            else:
                if isinstance(record_val, bytes):
                    record_val = record_val.decode()
                if isinstance(target_val, bytes):
                    target_val = target_val.decode()
                assert str(record_val) == str(target_val)


def assert_table(meta, table_name, target_columns):

    # Test that table exists
    tables = refresh_metadata(meta).tables
    assert table_name in tables
    assert tables[table_name].exists

    # Test that all columns were inserted
    test_table = tables[table_name]
    assert len(target_columns) == len(test_table.columns)
    assert all(c in test_table.columns for c in target_columns)


def make_table(meta, table_name, columns, col_types=None, data=None, dropfirst=True):

    exists = table_name in meta.tables and meta.tables[table_name].exists

    if exists and dropfirst:
        meta.tables[table_name].drop()

    if table_name in meta.tables:
        test_table = meta.tables[table_name]
    else:
        types = col_types or SITE_COL_TYPES
        test_table = Table(table_name, meta, *[Column(col, types.get(col, sqltypes.Text)) for col in columns])
        test_table.create()

    values = list(data.values()) if isinstance(data, dict) else data

    if values:
        with test_table.metadata.bind.connect() as conn:
            conn.execute(test_table.delete())
            conn.execute(test_table.insert(), values)

    return test_table


def refresh_metadata(meta):
    metadata = MetaData(meta.bind)
    metadata.reflect()
    return metadata


@pytest.fixture()
def db_settings():
    """ Remove/restore environment variables before/after implementing tests """

    db_env = os.environ.pop(ENVIRONMENT_VARIABLE, "")
    dj_env = os.environ.pop(DJANGO_SETTINGS_VAR, "")

    settings = conf.settings

    yield {
        ENVIRONMENT_VARIABLE: db_env,
        DJANGO_SETTINGS_VAR: dj_env,
    }

    if db_env:
        os.environ[ENVIRONMENT_VARIABLE] = db_env
    if dj_env:
        os.environ["DJANGO_SETTINGS_MODULE"] = dj_env
        reload_django_settings(settings)


def reload_django_settings(db_settings):
    try:
        db_settings._django_settings._wrapped = db_settings.empty
    except Exception:
        pass


def test_conf_settings_init(db_settings):

    db_env = db_settings[ENVIRONMENT_VARIABLE]
    dj_env = db_settings[DJANGO_SETTINGS_VAR]

    # Should not raise errors

    conf.settings._init_database_config()
    conf.settings._init_django_settings()
    conf.settings._init_database_info()

    # Test with invalid database configurations

    os.environ[ENVIRONMENT_VARIABLE] = db_env.replace("json", "txt")
    with pytest.raises(EnvironmentError, match="Invalid database configuration file"):
        conf.PgDatabaseSettings()

    os.environ[ENVIRONMENT_VARIABLE] = db_env.replace("config", "nope")
    with pytest.raises(EnvironmentError, match="Database configuration file does not exist"):
        conf.PgDatabaseSettings()

    os.environ[ENVIRONMENT_VARIABLE] = db_env.replace("test_config", "not")
    with pytest.raises(EnvironmentError, match="Database configuration file does not contain JSON"):
        conf.PgDatabaseSettings()

    os.environ[DJANGO_SETTINGS_VAR] = dj_env
    os.environ[ENVIRONMENT_VARIABLE] = db_env
    new_settings = conf.PgDatabaseSettings()

    if conf.DJANGO_INSTALLED:
        assert hasattr(new_settings._django_settings, "DATABASES")
    else:
        assert new_settings._django_settings == conf.EMPTY


def test_conf_settings_dbinfo(db_settings):

    db_env = db_settings[ENVIRONMENT_VARIABLE]
    dj_env = db_settings[DJANGO_SETTINGS_VAR]

    no_config_message = "No database configuration available"
    config_key_message = "Database configuration missing required key"
    invalid_dj_message = config_key_message if conf.DJANGO_INSTALLED else no_config_message

    # Test with no settings configured
    with pytest.raises(EnvironmentError, match=no_config_message):
        reload_django_settings(conf.settings)
        conf.PgDatabaseSettings().database_info

    # Test with no db config and empty Django settings
    os.environ[DJANGO_SETTINGS_VAR] = dj_env.replace("settings", "not_settings")
    with pytest.raises(EnvironmentError, match=no_config_message):
        reload_django_settings(conf.settings)
        conf.PgDatabaseSettings().database_info

    # Test with no db config and invalid Django settings
    os.environ[DJANGO_SETTINGS_VAR] = dj_env.replace("settings", "invalid_settings")
    with pytest.raises(EnvironmentError, match=invalid_dj_message):
        reload_django_settings(conf.settings)
        conf.PgDatabaseSettings().database_info

    # Test with invalid db config and Django settings
    os.environ[DJANGO_SETTINGS_VAR] = dj_env.replace("settings", "invalid_settings")
    os.environ[ENVIRONMENT_VARIABLE] = db_env.replace("test_config", "invalid_config")
    with pytest.raises(EnvironmentError, match=config_key_message):
        reload_django_settings(conf.settings)
        conf.PgDatabaseSettings().database_info

    if conf.DJANGO_INSTALLED:
        no_django_db_message = "No Django database configured for"

        # Test with invalid Django database in config and valid Django settings
        os.environ[DJANGO_SETTINGS_VAR] = dj_env
        os.environ[ENVIRONMENT_VARIABLE] = db_env.replace("test_config", "missing_django_db")
        with pytest.raises(EnvironmentError, match=no_django_db_message):
            reload_django_settings(conf.settings)
            conf.PgDatabaseSettings().database_info

        # Test with valid Django database in config and missing Django database
        os.environ[DJANGO_SETTINGS_VAR] = dj_env.replace("settings", "missing_django_db")
        os.environ[ENVIRONMENT_VARIABLE] = db_env
        with pytest.raises(EnvironmentError, match=no_django_db_message):
            reload_django_settings(conf.settings)
            conf.PgDatabaseSettings().database_info

        # Test with invalid Django database in config and missing Django database
        os.environ[DJANGO_SETTINGS_VAR] = dj_env.replace("settings", "missing_django_db")
        os.environ[ENVIRONMENT_VARIABLE] = db_env.replace("test_config", "missing_django_db")
        with pytest.raises(EnvironmentError, match=no_django_db_message):
            reload_django_settings(conf.settings)
            conf.PgDatabaseSettings().database_info

        # Test with custom values in config  and empty Django database
        os.environ[DJANGO_SETTINGS_VAR] = dj_env.replace("settings", "custom_databases")
        os.environ[ENVIRONMENT_VARIABLE] = db_env.replace("test_config", "custom_databases")
        reload_django_settings(conf.settings)

        assert not conf.PgDatabaseSettings().django_database
        assert conf.PgDatabaseSettings().host == "custom_host"
        assert conf.PgDatabaseSettings().password == "custom_pass"
        assert conf.PgDatabaseSettings().username == "custom_user"

    # Test with valid db config and Django settings

    os.environ[DJANGO_SETTINGS_VAR] = dj_env
    os.environ[ENVIRONMENT_VARIABLE] = db_env
    reload_django_settings(conf.settings)


def test_conf_settings_props():
    """ No need for db_settings fixture: environment variables need to be intact """

    django_installed = conf.DJANGO_INSTALLED

    # Test non-database properties

    test_settings = conf.PgDatabaseSettings()

    assert test_settings.connect_args == {"sslmode": "allow" if django_installed else "prefer"}
    assert test_settings.pooling_args == {"max_overflow": 10, "pool_recycle": -1, "pool_size": 5, "pool_timeout": 30}
    assert test_settings.django_db_key == "other"
    assert test_settings.date_format == DEFAULT_DATE_FORMAT
    assert test_settings.timestamp_format == DEFAULT_TIMESTAMP_FORMAT

    # Test database info dict

    test_settings = conf.PgDatabaseSettings()

    conf_engine = DEFAULT_ENGINE
    conf_name = "test_pg_database"
    conf_port = DEFAULT_PORT
    conf_host = None
    conf_user = "django" if django_installed else DEFAULT_USER
    conf_pass = None

    assert test_settings.database_info["drivername"] == conf_engine
    assert test_settings.database_info["database"] == conf_name
    assert test_settings.database_info["port"] == conf_port
    assert test_settings.database_info["host"] == conf_host
    assert test_settings.database_info["username"] == conf_user
    assert test_settings.database_info["password"] == conf_pass

    # Test top-level database properties

    test_settings = conf.PgDatabaseSettings()

    for prop in ("database_engine", "engine", "drivername"):
        assert getattr(test_settings, prop) == conf_engine
    for prop in ("database_name", "name", "database"):
        assert getattr(test_settings, prop) == conf_name
    for prop in ("database_port", "port"):
        assert getattr(test_settings, prop) == conf_port
    for prop in ("database_host", "host"):
        assert getattr(test_settings, prop) == conf_host
    for prop in ("database_user", "user", "username"):
        assert getattr(test_settings, prop) == conf_user
    for prop in ("database_password", "password"):
        assert getattr(test_settings, prop) == conf_pass

    # Test invalid top-level database properties (no dashes allowed)

    invalid_props = (
        "database-engine", "database-name", "database-port",
        "database-host", "database-user", "database-password",
        "database-nope", "database_nope",
        "django-nope", "django_nope", "NOPE",
    )
    for prop in invalid_props:
        assert getattr(test_settings, prop) is None

    # Test django database dict

    test_settings = conf.PgDatabaseSettings()

    django_engine = "django.contrib.gis.db.backends.postgis" if django_installed else None
    django_name = conf_name if django_installed else None
    django_port = None
    django_host = None
    django_user = "django" if django_installed else None
    django_pass = None
    django_options = {"sslmode": "allow"} if django_installed else None

    if not django_installed:
        assert test_settings.django_database == {}
    else:
        assert test_settings.django_database["django_engine"] == django_engine
        assert test_settings.django_database["django_name"] == conf_name
        assert test_settings.django_database["django_user"] == conf_user

    # Test django-specific database properties

    test_settings = conf.PgDatabaseSettings()

    for prop in ("ENGINE", "django_engine"):
        assert getattr(test_settings, prop) == django_engine
    for prop in ("NAME", "django_name"):
        assert getattr(test_settings, prop) == django_name
    for prop in ("PORT", "django_port"):
        assert getattr(test_settings, prop) == django_port
    for prop in ("HOST", "django_host"):
        assert getattr(test_settings, prop) == django_host
    for prop in ("USER", "django_user"):
        assert getattr(test_settings, prop) == django_user
    for prop in ("PASSWORD", "django_password"):
        assert getattr(test_settings, prop) == django_pass
    for prop in ("OPTIONS", "django_options"):
        assert getattr(test_settings, prop) == django_options

    # Test invalid django-specific database properties

    for prop in ("Engine", "Name", "Port", "Host", "User", "Password"):
        assert getattr(test_settings, prop) is None


def db_create(postgres_engine):

    database_exists = postgres_engine.execute(
        f"SELECT TRUE FROM pg_catalog.pg_database WHERE datname='{DATABASE_NAME}'"
    )
    if database_exists.scalar():
        return

    postgres_engine.execute(f"CREATE DATABASE {DATABASE_NAME} ENCODING 'utf8'")

    database_engine = create_engine(URL(
        drivername=DATABASE_INFO["drivername"],
        database=DATABASE_INFO["database"],
        username="postgres",
    ))
    with database_engine.connect() as conn:
        conn.execute("CREATE EXTENSION IF NOT EXISTS postgis;")
        conn.execute("CREATE EXTENSION IF NOT EXISTS postgis_topology;")


def db_drop(postgres_engine):

    with postgres_engine.connect() as conn:
        conn.execute(
            "SELECT pg_terminate_backend(pid) FROM pg_catalog.pg_stat_activity "
            f"WHERE pid <> pg_backend_pid() AND datname = '{DATABASE_NAME}'"
        )
        conn.execute(f"DROP DATABASE {DATABASE_NAME}")


@pytest.fixture(scope='session')
def db_engine():
    """ Create a database engine for use in implementing tests """

    postgres_engine = create_engine("postgresql:///postgres", isolation_level="AUTOCOMMIT")

    db_create(postgres_engine)
    engine = create_engine(URL(**DATABASE_INFO))

    yield engine

    engine.dispose()
    db_drop(postgres_engine)

    postgres_engine.dispose()


@pytest.fixture
def db_metadata(db_engine):
    """ Provides setup and teardown functionality for all database tests """

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message="Skipped unsupported reflection", category=SAWarning)

        # Setup: create data table with indexes, and return metadata

        meta = MetaData(db_engine)
        meta.reflect()

        make_table(meta, SITE_TABLE_NAME, SITE_TABLE_COLS, data=SITE_TEST_DATA)

        for col_name, col_type in GEOM_COL_TYPES.items():
            geom_sql = (
                f"ALTER TABLE {SITE_TABLE_NAME} "
                f"ALTER COLUMN {col_name} TYPE {col_type[0]} "
                f"USING {col_name}::{col_type[1]};"
            )
            with db_engine.connect() as conn:
                conn.execute(text(geom_sql).execution_options(autocommit=True))

        test_table = refresh_metadata(meta).tables[SITE_TABLE_NAME]

        # Create unique indexes on pk and obj_order
        Index("obj_index", "pk", unique=True, _table=test_table).create()
        Index("order_index", "obj_order", unique=True, _table=test_table).create()
        Index("hash_index", "obj_hash", unique=True, _table=test_table).create()

        # Create search index on primary test data columns
        expressions = (func.to_tsvector("english", text("||' '||".join(SEARCHABLE_COLS))),)
        Index("tsvector_index", *expressions, postgresql_using="gin", _table=test_table).create()

        yield test_table.metadata

        # Teardown: drop all data related tables used in the tests

        meta = refresh_metadata(meta)
        tables = (t for t in meta.tables.values() if t.name in ALL_TEST_TABLES)
        for fk in (c for t in tables for c in t.foreign_key_constraints):
            DropConstraint(fk, cascade=True).execute(meta.bind)

        meta = refresh_metadata(meta)
        for test_table in (t for t in meta.tables.values() if t.name in ALL_TEST_TABLES):
            test_table.drop()


def test_column_types():

    def assert_column_type(test_type, col_val, str_val=None, type_to_str_val=None):

        lower = test_type.lower()
        upper = test_type.upper()
        test_type = ''.join(random.choice(c) for c in zip(upper, lower))

        str_type = types.type_to_string(test_type)
        col_type = types.column_type_for(test_type)

        assert str_type == (str_val or lower), f"Wrong type_to_string for {test_type}"
        assert col_type == col_val, f"Wrong column_type_for {test_type}"
        assert col_type == types.COLUMN_TYPE_MAP[str_type], f"Wrong column type mapping for {test_type}"
        assert types.type_to_string(col_type) == (type_to_str_val or str_type), (
            f"Wrong type_to_string return conversion for {test_type}"
        )

        assert validation.validate_column_type(str_type) == str_type, f"String validation failed for {test_type}"
        assert validation.validate_column_type(col_type) == col_type, f"Type validation failed for {test_type}"

    # Test validation for invalid column type strings

    inject_sql = "DROP USER 'postgres' IF EXISTS"

    # Test validate_column_type with empty types
    with pytest.raises(ValueError, match="Invalid column type: empty"):
        validation.validate_column_type(None)
    with pytest.raises(ValueError, match="Invalid column type: empty"):
        validation.validate_column_type("")
    # Test validate_column_type with simple and complex string types
    with pytest.raises(ValueError, match="Invalid column type"):
        validation.validate_column_type(inject_sql)
    with pytest.raises(ValueError, match="Invalid column type"):
        validation.validate_column_type("nope(POINT,4326)")
    with pytest.raises(ValueError, match="Invalid column type"):
        validation.validate_column_type(f"geometry({inject_sql})")
    # Test validate_column_type with non-string types
    with pytest.raises(ValueError, match="Invalid column type"):
        validation.validate_column_type(["not", "a", "sqlalchemy.sql.sqltypes.TypeEngine"])

    # Test all supported column type mappings and validation

    # Boolean related
    assert_column_type("BOOL", sqltypes.Boolean, type_to_str_val="boolean")
    assert_column_type("BOOLEAN", sqltypes.Boolean)
    # Number related
    assert_column_type("BIGINT", sqltypes.BigInteger)
    assert_column_type("INT", sqltypes.Integer, type_to_str_val="integer")
    assert_column_type("INTEGER", sqltypes.Integer)
    assert_column_type("DECIMAL", sqltypes.Numeric, type_to_str_val="numeric")
    assert_column_type("DOUBLE", sqltypes.Numeric, "numeric")
    assert_column_type("FLOAT", sqltypes.Float)
    assert_column_type("NUMERIC", sqltypes.Numeric)
    assert_column_type("NUMBER", sqltypes.Numeric, "numeric")
    # Date related
    assert_column_type("DATE", sqltypes.Date)
    assert_column_type("DATETIME", sqltypes.DateTime, "timestamp")
    assert_column_type("TIMESTAMP", sqltypes.DateTime)
    # Text related
    assert_column_type("BYTEA", sqltypes.LargeBinary)
    assert_column_type("BINARY", sqltypes.LargeBinary, "bytea")
    assert_column_type("STRING", sqltypes.UnicodeText, "text")
    assert_column_type("TEXT", sqltypes.UnicodeText)
    assert_column_type("UNICODE", sqltypes.UnicodeText, "text")
    assert_column_type("VARCHAR", sqltypes.UnicodeText, type_to_str_val="text")
    # GIS and JSON related
    assert_column_type("GEOMETRY", gistypes.Geometry)
    assert_column_type("GEOGRAPHY", gistypes.Geography)
    assert_column_type("RASTER", gistypes.Raster)
    assert_column_type("JSON", postgresql.json.JSON)
    assert_column_type("JSONB", postgresql.json.JSONB)


def test_pooling_params():

    # Test with invalid params

    with pytest.raises(ValueError, match="Invalid pooling params"):
        validation.validate_pooling_params(["not", "a", "dict"])
    with pytest.raises(ValueError, match="Invalid pooling params"):
        validation.validate_pooling_params({"not": "supported"})
    with pytest.raises(ValueError, match="Pooling params require integer values"):
        validation.validate_pooling_params({"pool_size": 20, "max_overflow": "nope"})

    # Test with empty params provided

    assert validation.validate_pooling_params(None) == {}
    assert validation.validate_pooling_params({}) == {}

    # Test with all supported params provided

    pooling_params = {"max_overflow": 0, "pool_recycle": 60, "pool_size": 20, "pool_timeout": 30}
    assert validation.validate_pooling_params(pooling_params) == pooling_params


def test_database_setup(db_metadata):
    database_query = db_metadata.bind.execute(
        "SELECT TRUE FROM pg_catalog.pg_database "
        f"WHERE datname='{DATABASE_NAME}'"
    )
    assert database_query.scalar()


def test_create_index(db_metadata):
    table_name = SITE_TABLE_NAME
    test_table = db_metadata.tables[table_name]

    inject_sql = "DROP USER 'postgres' IF EXISTS"

    # Test with invalid parameters

    with pytest.raises(ValueError, match="No table named"):
        schema.create_index("nope", "site_addr")
    with pytest.raises(ValueError, match="Invalid index column names"):
        schema.create_index(table_name, [])
    with pytest.raises(ValueError, match="Invalid index column names"):
        schema.create_index(table_name, ["site_addr", "nope"])
    with pytest.raises(ValueError, match="Invalid index column names"):
        schema.create_index(table_name, ["site_addr", inject_sql])
    with pytest.raises(ValueError, match="Invalid index operation for multiple columns"):
        schema.create_index(table_name, "test_json,test_none", None, "json_full")
    with pytest.raises(ValueError, match="Invalid index operation for multiple columns"):
        schema.create_index(table_name, "test_json,test_none", None, "json_path")
    with pytest.raises(ValueError, match="Invalid index operation for multiple columns"):
        schema.create_index(table_name, "site_addr,test_none", None, "spatial")
    with pytest.raises(ValueError, match="Invalid column type for spatial index"):
        schema.create_index(table_name, "site_addr", None, "spatial")
    with pytest.raises(ValueError, match="Unsupported index type"):
        schema.create_index(table_name, "site_addr", None, "nope")

    # Test default index creation
    column_name = "site_addr"
    schema.create_index(table_name, column_name)
    assert_index(db_metadata, table_name, column_name, index_name=f"{table_name}_{column_name}_idx")

    # Test unique index creation with overridden index name
    column_name = "obj_order"
    schema.create_index(table_name, column_name, f"{table_name}_unique_idx", index_op="unique")
    assert_index(db_metadata, table_name, column_name, index_name=f"{table_name}_unique_idx")

    # Test spatial index creation on point column
    column_name = "test_geom"
    schema.create_index(table_name, column_name, index_op="spatial")
    assert_index(db_metadata, table_name, column_name, index_name=f"{table_name}_{column_name}_spatial_idx")

    # Test spatial index creation on line column
    column_name = "test_line"
    schema.create_index(table_name, column_name, index_op="spatial")
    assert_index(db_metadata, table_name, column_name, index_name=f"{table_name}_{column_name}_spatial_idx")

    # Test spatial index creation on polygon column
    column_name = "test_poly"
    schema.create_index(table_name, column_name, index_op="spatial")
    assert_index(db_metadata, table_name, column_name, index_name=f"{table_name}_{column_name}_spatial_idx")

    # Test json index creation (all ops)
    column_name = "test_json"
    schema.create_index(table_name, column_name, index_op="json_full")
    assert_index(db_metadata, table_name, column_name, index_name=f"{table_name}_{column_name}_json_full_idx")

    # Test json index creation (path ops)
    column_name = "test_json"
    schema.create_index(table_name, column_name, f"{table_name}_json_path_idx", index_op="json_path")
    assert_index(db_metadata, table_name, column_name, index_name=f"{table_name}_json_path_idx")

    # Test multi-column index creation on in-memory table

    # TSVECTOR indexes are not loaded from database via table reflection
    schema.create_index(test_table, SEARCHABLE_COLS, f"{table_name}_search_idx", index_op="to_tsvector")
    assert_index(db_metadata, table_name, index_name=f"{table_name}_search_idx")

    # COALESCE indexes are not loaded from database via table reflection
    column_names = [c for c in SEARCHABLE_COLS if c.startswith("site")]
    index_cols = "_".join(col for col in column_names)
    index_name = f"{table_name}_{index_cols}_coalesce_idx"
    schema.create_index(test_table, column_names, index_op="coalesce")
    assert_index(db_metadata, table_name, index_name=index_name)


def test_drop_index(db_metadata):
    table_name = SITE_TABLE_NAME
    test_table = db_metadata.tables[table_name]

    # Test with invalid parameters

    # Should raise no errors
    schema.drop_index("nope", column_names="site_addr", ignore_errors=True)
    schema.drop_index(table_name, ignore_errors=True)
    schema.drop_index(table_name, index_name="nope", ignore_errors=True)

    with pytest.raises(ValueError, match="No table named"):
        schema.drop_index("nope", column_names="site_addr", ignore_errors=False)
    with pytest.raises(ValueError, match="No index name provided"):
        schema.drop_index(table_name, ignore_errors=False)
    with pytest.raises(Exception):
        schema.drop_index(table_name, index_name="nope", ignore_errors=False)

    # Test removing APN index created during setup
    index_name = "order_index"
    schema.drop_index(test_table, index_name, ignore_errors=False)
    assert_index(db_metadata, table_name, index_name=index_name, exists=False)

    # Test removing search index created during setup
    index_name = "tsvector_index"
    schema.drop_index(table_name, index_name, ignore_errors=False)
    assert_index(db_metadata, table_name, index_name=index_name, exists=False)

    # Test removing indexes by column name

    pk_index = f"{table_name}_obj_index_obj_hash_idx"

    Index(pk_index, "pk", unique=True, _table=test_table).create()
    assert_index(db_metadata, table_name, index_name=pk_index)
    schema.drop_index(table_name, column_names="obj_index,obj_hash", ignore_errors=False)
    assert_index(db_metadata, table_name, index_name=pk_index, exists=False)

    Index(pk_index, "pk", unique=True, _table=test_table).create()
    assert_index(db_metadata, table_name, index_name=pk_index)
    schema.drop_index(table_name, column_names=["obj_index", "obj_hash"], ignore_errors=False)
    assert_index(db_metadata, table_name, index_name=pk_index, exists=False)


def test_has_index(db_metadata):
    table_name = SITE_TABLE_NAME
    test_table = db_metadata.tables[table_name]

    # Test indexes created during setup
    assert schema.has_index(test_table, index_name="order_index")
    assert schema.has_index(table_name, index_name="tsvector_index")

    # Test that indexes no longer exist after being dropped

    Index("order_index", _table=test_table).drop()
    Index("tsvector_index", _table=test_table).drop()

    assert not schema.has_index(test_table, index_name="order_index")
    assert not schema.has_index(table_name, index_name="tsvector_index")


def test_create_table(db_metadata):

    table_name = SITE_TABLE_NAME
    inject_sql = "DROP USER 'postgres' IF EXISTS"

    # Test with invalid parameters

    tmp_table_name = ALL_TEST_TABLES[2]

    # Test invalid table names
    with pytest.raises(ValueError, match="No table name specified"):
        schema.create_table(None, col="varchar")
    with pytest.raises(ValueError, match="Invalid table name"):
        schema.create_table(inject_sql, column_one="varchar")
    with pytest.raises(ValueError, match="Table already exists"):
        schema.create_table(table_name, dropfirst=False, **SITE_COL_TYPES)
    # Test invalid index columns
    with pytest.raises(ValueError, match="No column names specified"):
        schema.create_table("no_columns")
    with pytest.raises(ValueError, match="Invalid index column names"):
        schema.create_table(tmp_table_name, index_cols="col,nope", col="varchar", dropfirst=True)
    with pytest.raises(ValueError, match="Invalid index column names"):
        schema.create_table(tmp_table_name, index_cols=["col", "nope"], col="varchar", dropfirst=True)
    with pytest.raises(ValueError, match="Invalid index column names"):
        schema.create_table(
            tmp_table_name, index_cols={"col": "unique", "nope": "unique"}, col="varchar", dropfirst=True
        )

    column_types = dict(SITE_COL_TYPES)
    column_types["_engine"] = postgresql.json.JSON
    target_types = dict(column_types)
    target_types["engine"] = target_types.pop("_engine")

    # Test with no indexes and site columns

    schema.create_table(table_name, dropfirst=True, **column_types)
    assert_table(db_metadata, table_name, target_types)
    assert not refresh_metadata(db_metadata).tables[table_name].indexes

    # Test with no indexes, with engine provided

    refresh_metadata(db_metadata).tables[table_name].drop()

    schema.create_table(table_name, dropfirst=True, engine=db_metadata.bind, **column_types)
    assert_table(db_metadata, table_name, target_types)
    assert not refresh_metadata(db_metadata).tables[table_name].indexes

    # Test with string of index columns

    refresh_metadata(db_metadata).tables[table_name].drop()

    schema.create_table(table_name, index_cols="pk,obj_order", **column_types)
    assert_table(db_metadata, table_name, target_types)
    assert_index(db_metadata, table_name, index_name=f"{table_name}_pk_obj_order_idx")

    # Test with list of index columns and column type names

    refresh_metadata(db_metadata).tables[table_name].drop()

    schema.create_table(
        table_name,
        index_cols=["pk", "obj_order"],
        dropfirst=False,
        **{k: types.type_to_string(v) for k, v in column_types.items()}
    )
    assert_table(db_metadata, table_name, target_types)
    assert_index(db_metadata, table_name, index_name=f"{table_name}_pk_idx")
    assert_index(db_metadata, table_name, index_name=f"{table_name}_obj_order_idx")

    # Test with dict of index columns and column type instances

    refresh_metadata(db_metadata).tables[table_name].drop()

    schema.create_table(
        table_name,
        index_cols={"pk": "unique", "obj_order": "unique", "pk,obj_order": "unique"},
        dropfirst=False,
        **{k: v() for k, v in column_types.items()}
    )
    assert_table(db_metadata, table_name, target_types)
    assert_index(db_metadata, table_name, index_name=f"{table_name}_pk_unique_idx")
    assert_index(db_metadata, table_name, index_name=f"{table_name}_obj_order_unique_idx")
    assert_index(db_metadata, table_name, index_name=f"{table_name}_pk_obj_order_unique_idx")


def test_drop_table(db_metadata):

    site_table = db_metadata.tables[SITE_TABLE_NAME]

    schema.drop_table(SITE_TABLE_NAME)
    assert SITE_TABLE_NAME not in refresh_metadata(db_metadata).tables

    # Test checkfirst param

    schema.drop_table(site_table)
    with pytest.raises(ValueError, match="No table named"):
        schema.drop_table(site_table, checkfirst=False)

    schema.drop_table(site_table.name)
    with pytest.raises(ValueError, match="No table named"):
        schema.drop_table(site_table.name, checkfirst=False)


def test_get_engine(db_metadata):

    # Test with invalid params

    with pytest.raises(ValueError, match="Invalid pooling params"):
        schema.get_engine(pooling_args=["not", "a", "dict"])
    with pytest.raises(ValueError, match="Invalid pooling params"):
        schema.get_engine(pooling_args={"not": "supported"})
    with pytest.raises(ValueError, match="Pooling params require integer values"):
        schema.get_engine(pooling_args={"pool_size": 20, "max_overflow": "nope"})

    # Test with default params

    test_metadata = MetaData(schema.get_engine())
    test_metadata.reflect()
    assert sorted(test_metadata.tables) == sorted(db_metadata.tables)

    # Test with empty params provided

    test_metadata = MetaData(schema.get_engine(connect_args={}, pooling_args={}))
    test_metadata.reflect()
    assert sorted(test_metadata.tables) == sorted(db_metadata.tables)

    # Test with all params provided

    test_metadata = MetaData(schema.get_engine(
        connect_args={"sslmode": "require"},
        pooling_args={"pool_size": 20, "max_overflow": 0},
    ))
    test_metadata.reflect()
    assert sorted(test_metadata.tables) == sorted(db_metadata.tables)


def test_get_metadata(db_metadata):

    # Test with invalid params

    with pytest.raises(ValueError, match="Invalid pooling params"):
        schema.get_metadata(pooling_args=["not", "a", "dict"])
    with pytest.raises(ValueError, match="Invalid pooling params"):
        schema.get_metadata(pooling_args={"not": "supported"})
    with pytest.raises(ValueError, match="Pooling params require integer values"):
        schema.get_metadata(pooling_args={"pool_size": 20, "max_overflow": "nope"})

    # Test with default params

    test_metadata = schema.get_metadata()
    assert sorted(test_metadata.tables) == sorted(db_metadata.tables)

    # Test with empty params provided

    test_metadata = schema.get_metadata(connect_args={}, pooling_args={})
    assert sorted(test_metadata.tables) == sorted(db_metadata.tables)

    # Test with all params provided

    test_metadata = schema.get_metadata(
        connect_args={"sslmode": "require"},
        pooling_args={"pool_size": 20, "max_overflow": 0},
    )
    assert sorted(test_metadata.tables) == sorted(db_metadata.tables)


def test_get_table(db_metadata):

    schema.get_table("nope", ignore_missing=True)
    with pytest.raises(ValueError, match="No table named"):
        schema.get_table("nope")

    # Test for existance of custom table
    assert schema.get_table(SITE_TABLE_NAME).exists
    assert schema.get_table(POSTGIS_TABLE_NAME, engine=db_metadata.bind).exists


def test_get_table_count(db_metadata):

    target_count = len(SITE_TEST_DATA)

    with pytest.raises(ValueError, match="No table named"):
        schema.get_table_count("nope")

    # Test count for initialized data

    assert schema.get_table_count(SITE_TABLE_NAME) == target_count
    assert schema.get_table_count(db_metadata.tables[SITE_TABLE_NAME]) == target_count


def test_get_tables(db_metadata):

    all_tables = schema.get_tables()
    table_name = SITE_TABLE_NAME

    # Test for existance of custom table
    assert table_name in all_tables
    assert all_tables[table_name].exists

    # Test for existance of tables specified as string

    some_tables = schema.get_tables(f"{table_name},nope")

    assert len(some_tables) == 1
    assert table_name in some_tables
    assert some_tables[table_name].exists

    # Test for existance of tables specified as list

    some_tables = schema.get_tables(["nope", f"{table_name}"])

    assert len(some_tables) == 1
    assert table_name in some_tables
    assert some_tables[table_name].exists

    # Test for existance of multiple tables when engine is provided

    some_tables = schema.get_tables((f"{POSTGIS_TABLE_NAME}", f"{table_name}"), engine=db_metadata.bind)

    assert len(some_tables) == 2
    assert table_name in some_tables
    assert POSTGIS_TABLE_NAME in some_tables
    assert some_tables[table_name].exists


def test_insert_from(db_metadata):

    from_table_name = SITE_TABLE_NAME
    into_table_name = ALL_TEST_TABLES[1]

    # Test with invalid parameters

    with pytest.raises(ValueError, match="No table named"):
        sql.insert_from("nope", SITE_TABLE_NAME, "*")
    with pytest.raises(ValueError, match="No table named"):
        sql.insert_from(from_table_name, "nope", "*")
    with pytest.raises(ValueError, match="Join columns missing in.*table"):
        sql.insert_from(from_table_name, SITE_TABLE_NAME, "*", join_columns="nope")

    # Test with asterisk to insert all columns

    # Test that table is created with all columns if it doesn't exist
    sql.insert_from(from_table_name, into_table_name, "*", create_if_not_exists=True)
    assert_table(db_metadata, into_table_name, SITE_TABLE_COLS)
    assert_records(db_metadata, into_table_name, SITE_DATA_DICT)

    # Test that table is updated with new records if it does
    sql.insert_from(from_table_name, into_table_name, "*", create_if_not_exists=False)
    assert_records(db_metadata, into_table_name, SITE_DATA_DICT, target_count=(len(SITE_TEST_DATA) * 2))

    # Test with subset of columns

    target_columns = ("pk", "obj_order", "obj_hash", "site_zip")

    refresh_metadata(db_metadata).tables[into_table_name].drop()

    # Test that table is not created when only invalid column names are provided
    sql.insert_from(from_table_name, into_table_name, "ignore,these", create_if_not_exists=True)
    assert into_table_name not in refresh_metadata(db_metadata).tables

    # Test that table is created with subset of columns if it doesn't exist, withe engine provided
    sql.insert_from(
        from_table_name, into_table_name, target_columns, create_if_not_exists=True, engine=db_metadata.bind
    )
    assert_table(db_metadata, into_table_name, target_columns)
    assert_records(db_metadata, into_table_name, SITE_DATA_DICT)

    # Test that table has no new records when only invalid column names are provided
    sql.insert_from(from_table_name, into_table_name, "ignore,these")
    assert_table(db_metadata, into_table_name, target_columns)
    assert_records(db_metadata, into_table_name, SITE_DATA_DICT)

    # Test that existing table is updated with new records, ignoring two invalid columns
    target_count = len(SITE_DATA_DICT) * 2
    sql.insert_from(from_table_name, into_table_name, target_columns + ("ignore", "these"))
    assert_records(db_metadata, into_table_name, SITE_DATA_DICT, target_count=target_count)

    # Test that table is not affected by duplicate records
    into_table = refresh_metadata(db_metadata).tables[into_table_name]
    target_count = select([func.count()]).select_from(into_table).scalar()
    sql.insert_from(from_table_name, into_table_name, target_columns, join_columns=["pk", "obj_order", "obj_hash"])
    assert_records(db_metadata, into_table_name, SITE_DATA_DICT, target_count=target_count)


def test_insert_into(db_metadata):

    target_cols = ["pk", "str", "bool", "int"]
    into_table_name = ALL_TEST_TABLES[1]

    # Test with invalid parameters

    make_table(db_metadata, into_table_name, target_cols)

    with pytest.raises(ValueError, match="No table named"):
        sql.insert_into("nope", [("one",), ("two",), ("three",)], "str")
    with pytest.raises(ValueError, match="Invalid column names"):
        sql.insert_into(into_table_name, [("one",), ("two",), ("three",)], "no,nope,wrong")

    with pytest.raises(ValueError, match="Values provided do not match columns"):
        sql.insert_into(into_table_name, [("one", "two"), ("three",), ("four",)], "str")
    with pytest.raises(ValueError, match="Values provided do not match columns"):
        sql.insert_into(into_table_name, [("one",), ("two", "three"), ("four",)], "str")
    with pytest.raises(ValueError, match="Values provided do not match columns"):
        sql.insert_into(into_table_name, ["one", ("two",), ("three", "four")], "str")

    # Test with empty values (logs warning and exits)
    sql.insert_into(into_table_name, [], "str")

    # Cleanup in preparation for data tests
    refresh_metadata(db_metadata).tables[into_table_name].drop()

    # Prepare target data for insert

    target_vals = [
        {"pk": "128_2761_66", "str": "535 STATE ST", "bool": 1, "int": 42},
        {"pk": "128_2761_67", "str": "535 STATE ST", "bool": 0, "int": 86},
        {"pk": "128_2761_68", "str": "535 STATE ST", "bool": 1, "int": -4},
    ]
    target_data = {p["pk"]: p for p in target_vals}

    # Ensure the order of values matches the target columns
    insert_vals = [[p[c] for c in target_cols] for p in target_vals]

    # Test insert with new table and no types specified

    sql.insert_into(into_table_name, insert_vals, target_cols, create_if_not_exists=True)
    assert_table(db_metadata, into_table_name, target_cols)
    assert_records(db_metadata, into_table_name, target_data, target_cols)

    for inserted in refresh_metadata(db_metadata).tables[into_table_name].columns:
        assert str(inserted.type).lower() == "text"

    # Test insert duplicate records into existing table, with engine provided

    sql.insert_into(into_table_name, insert_vals, target_cols, engine=db_metadata.bind)
    target_count = len(target_vals) * 2
    assert_table(db_metadata, into_table_name, target_cols)
    assert_records(db_metadata, into_table_name, target_data, target_cols, target_count)


def test_select_from(db_metadata):

    from_table_name = SITE_TABLE_NAME
    into_table_name = ALL_TEST_TABLES[1]

    # Test with invalid parameters

    with pytest.raises(ValueError, match="No table named"):
        sql.select_from("nope", into_table_name, "*")
    with pytest.raises(ValueError, match="Table.*already exists"):
        sql.select_from(from_table_name, SITE_TABLE_NAME, "*")

    # Test with invalid columns
    sql.select_from(from_table_name, into_table_name, "ignore,these")
    assert into_table_name not in refresh_metadata(db_metadata).tables

    # Test with asterisk to insert all columns, with engine provided
    sql.select_from(from_table_name, into_table_name, "*", engine=db_metadata.bind)
    assert_table(db_metadata, into_table_name, SITE_TABLE_COLS)
    assert_records(db_metadata, into_table_name, SITE_DATA_DICT)

    # Test with subset of columns (including two invalid)

    target_columns = ("pk", "obj_order", "obj_hash", "site_zip")

    refresh_metadata(db_metadata).tables[into_table_name].drop()

    sql.select_from(from_table_name, into_table_name, target_columns + ("ignore", "these"))
    assert_table(db_metadata, into_table_name, target_columns)
    assert_records(db_metadata, into_table_name, SITE_DATA_DICT, target_columns=target_columns)


def test_select_into(db_metadata):

    into_table_name = ALL_TEST_TABLES[1]

    # Test with invalid parameters

    inject_sql = "DROP USER 'postgres' IF EXISTS"

    with pytest.raises(ValueError, match="Table.*already exists"):
        sql.select_into(SITE_TABLE_NAME, [("one",), ("two",), ("three",)], "val")
    with pytest.raises(ValueError, match="No columns to select"):
        sql.select_into(into_table_name, [("one",), ("two",), ("three",)], "")
    with pytest.raises(ValueError, match="Invalid column names"):
        sql.select_into(into_table_name, [("one",), ("two",), ("three",)], inject_sql)
    with pytest.raises(ValueError, match="Column types provided do not match columns"):
        sql.select_into(into_table_name, [("one",), ("two",), ("three",)], "val", "bool,json")

    with pytest.raises(ValueError, match="Values provided do not match columns"):
        sql.select_into(into_table_name, [("one", "two"), ("three",), ("four",)], "val")
    with pytest.raises(ValueError, match="Values provided do not match columns"):
        sql.select_into(into_table_name, [("one",), ("two", "three"), ("four",)], "val")
    with pytest.raises(ValueError, match="Values provided do not match columns"):
        sql.select_into(into_table_name, ["one", ("two",), ("three", "four")], "val")

    # Test with empty values (logs warning and exits)
    sql.select_into(into_table_name, [], "val")

    # Prepare target data for insert

    target_cols = SITE_TABLE_COLS
    target_data = {
        p["pk"]: {
            # Stringify JSON values for text comparison after insertion
            k: json.dumps(v) if "json" in k else v
            for k, v in p.items()
        }
        for p in SITE_TEST_DATA
    }

    # Ensure the order of values matches the target columns
    insert_vals = [[p[c] for c in target_cols] for p in target_data.values()]
    insert_types = [
        # Use existing type dict to derive type names for each column
        SITE_COL_TYPES.get(col, sqltypes.Text).__visit_name__ for col in target_cols
    ]

    # Test insert with no types specified

    sql.select_into(into_table_name, insert_vals, ",".join(target_cols))
    assert_table(db_metadata, into_table_name, target_cols)
    assert_records(db_metadata, into_table_name, target_data, target_cols)

    for inserted in refresh_metadata(db_metadata).tables[into_table_name].columns:
        assert str(inserted.type).lower() == "text"

    # Test insert with specific type names, with engine provided

    refresh_metadata(db_metadata).tables[into_table_name].drop()

    target_data = {p["pk"]: p for p in SITE_TEST_DATA}

    sql.select_into(into_table_name, insert_vals, target_cols, insert_types, engine=db_metadata.bind)
    assert_table(db_metadata, into_table_name, target_cols)
    assert_records(db_metadata, into_table_name, target_data, target_cols)

    for idx, inserted in enumerate(refresh_metadata(db_metadata).tables[into_table_name].columns):
        assert str(inserted.type).lower() == insert_types[idx].lower()


def test_update_from(db_metadata):

    from_table_name = SITE_TABLE_NAME
    into_table_name = ALL_TEST_TABLES[1]

    # Test with invalid parameters

    with pytest.raises(ValueError, match="No table named"):
        sql.update_from("nope", SITE_TABLE_NAME, "pk")
    with pytest.raises(ValueError, match="No table named"):
        sql.update_from(from_table_name, "nope", "pk")
    with pytest.raises(ValueError, match="Join columns missing"):
        sql.update_from(from_table_name, SITE_TABLE_NAME, "")
    with pytest.raises(ValueError, match="Join columns missing"):
        sql.update_from(from_table_name, SITE_TABLE_NAME, ["site_addr", "nope"])

    # Create and populate a table with mangled data for testing updates

    orig_data = SITE_DATA_DICT
    join_cols = tuple(c for c in SITE_COL_TYPES if not c.startswith("test"))
    test_cols = tuple(SITE_COL_TYPES)
    test_data = {
        pk: {
            # Reverse string values for all but test columns
            k: v if k in test_cols else str(v)[::-1]
            for k, v in record.items()
        }
        for pk, record in orig_data.items()
    }

    make_table(db_metadata, into_table_name, SITE_TABLE_COLS, data=test_data)
    assert_table(db_metadata, into_table_name, SITE_TABLE_COLS)
    assert_records(db_metadata, into_table_name, test_data)

    # Test that all records are updated by join on target columns, with engine provided
    sql.update_from(from_table_name, into_table_name, ",".join(join_cols), engine=db_metadata.bind)
    assert_records(db_metadata, into_table_name, orig_data)

    # Reset the updated table to mangled values for next test
    make_table(db_metadata, into_table_name, SITE_TABLE_COLS, data=test_data, dropfirst=False)
    assert_records(db_metadata, into_table_name, test_data)

    # Test that no records are updated when only invalid column names are provided
    sql.update_from(from_table_name, into_table_name, join_cols, "ignore,these")
    assert_records(db_metadata, into_table_name, test_data)

    # Test that only select fields are updated by join on target columns, and invalid columns are ignored

    update_cols = set(c for c in SITE_TABLE_COLS if c.startswith("site"))
    ignore_cols = set(SITE_TABLE_COLS).difference(update_cols)

    sql.update_from(from_table_name, into_table_name, join_cols, update_cols.union(("ignore", "these")))
    assert_records(db_metadata, into_table_name, orig_data, target_columns=update_cols)
    assert_records(db_metadata, into_table_name, test_data, target_columns=ignore_cols)


def test_update_rows(db_metadata):

    def update_zips(row):
        """
        Add integer zip to new column and json
        Row: pk, site_zip, size_zip_num, site_json
        """
        row = list(row)
        row[3]["zip"] = row[2] = int(row[1])
        return row

    def update_some(row):
        """
        Add "changed" key to JSON depending on address
        Row: pk, site_zip, size_zip_num, site_json
        """
        row = list(row)
        if row[-1]["number"] != "7350":
            row = None
        else:
            row[-1]["changed"] = True
        return row

    def update_none(row):
        """ Skips updates to all rows """
        return None

    test_table_name = ALL_TEST_TABLES[1]
    empty_table_name = ALL_TEST_TABLES[2]

    # Test with invalid parameters

    with pytest.raises(ValueError, match="No table named"):
        sql.update_rows("nope", "pk", "site_addr", update_zips)
    with pytest.raises(ValueError, match="Join columns missing"):
        sql.update_rows(SITE_TABLE_NAME, "", "site_addr", update_zips)
    with pytest.raises(ValueError, match="Target columns missing"):
        sql.update_rows(SITE_TABLE_NAME, "pk", "", update_zips)
    with pytest.raises(ValueError, match="Invalid update function"):
        sql.update_rows(SITE_TABLE_NAME, "pk", "site_addr", None, 0)
    with pytest.raises(ValueError, match="Invalid batch size"):
        sql.update_rows(SITE_TABLE_NAME, "pk", "site_addr", update_zips, 0)

    with pytest.raises(ValueError, match="Join columns missing in.*table"):
        sql.update_rows(SITE_TABLE_NAME, "pk,nope", "site_addr", update_zips)
    with pytest.raises(ValueError, match="Target columns missing in.*table"):
        sql.update_rows(SITE_TABLE_NAME, "pk", ["site_addr", "nope"], update_zips)

    target_cols = ("site_zip", "site_zip_num", "site_json")

    # Test when there are no rows to update

    make_table(db_metadata, empty_table_name, SITE_TABLE_COLS)
    sql.update_rows(empty_table_name, ["pk"], SITE_TABLE_COLS, update_zips)

    # Prepare a fully populated table

    # Insert test table columns in undefined order via `set`
    test_cols = list(set(SITE_TABLE_COLS).union(target_cols))[::-1]
    test_data = dict(SITE_DATA_DICT)
    # New table will have new column with numeric zip
    test_types = dict(SITE_COL_TYPES)
    test_types["site_zip_num"] = sqltypes.Integer

    # Create a test table with default data
    make_table(db_metadata, test_table_name, test_cols, test_types, data=test_data)
    assert_table(db_metadata, test_table_name, test_cols)
    assert_records(db_metadata, test_table_name, test_data)

    # Test updating all the rows in data table, with engine provided

    sql.update_rows(test_table_name, "pk", target_cols, update_zips, batch_size=3, engine=db_metadata.bind)

    target_data = {}
    for pk, record in copy.deepcopy(test_data).items():
        parsed = int(record["site_zip"])
        record["site_zip_num"] = parsed
        record["site_json"]["zip"] = parsed
        target_data[pk] = record

    assert_records(db_metadata, test_table_name, target_data, target_cols)
    assert target_data != test_data
    assert f"tmp_{test_table_name}" not in refresh_metadata(db_metadata).tables

    # Test updating some the rows depending on address, joining on more than one column

    sql.update_rows(test_table_name, "pk,obj_order", target_cols, update_some, batch_size=3)

    test_data = target_data
    target_data = {}

    for pk, record in copy.deepcopy(test_data).items():
        if record["site_json"]["number"] == "7350":
            record["site_json"]["changed"] = True
        target_data[record["pk"]] = record

    assert_records(db_metadata, test_table_name, target_data, target_cols)
    assert target_data != test_data
    assert f"tmp_{test_table_name}" not in refresh_metadata(db_metadata).tables

    # Test updating none of the rows in the data table

    sql.update_rows(test_table_name, ["pk", "obj_order"], target_cols, update_none, batch_size=3)

    assert_records(db_metadata, test_table_name, target_data, target_cols)
    assert f"tmp_{test_table_name}" not in refresh_metadata(db_metadata).tables


def test_values_clause(db_metadata):

    into_table_name = ALL_TEST_TABLES[1]

    # Define VALUES clause columns, types and values

    target_cols = ["pk", "hash", "int", "str", "bool", "float"]
    target_types = ["text", "binary", "integer", "unicode", "boolean", "float"]
    target_values = [
        ("128_2761_66", b"0065455eeaf0bd713f26dfb95f5dd9bf", 535, "STATE ST", 1, 42.5),
        ("128_2761_67", b"786334b2a927a8e3c6eee727849ab316", 536, "STATE ST", 0, "86.7"),
        ("128_2761_68", b"c47fa34a1d7a1f8bd686e98c201d12bd", "537", "STATE ST", True, -4),
    ]
    target_data = {p["pk"]: p for p in [dict(zip(target_cols, row)) for row in target_values]}

    # Test invalid types in VALUES clause

    with pytest.raises(ArgumentError, match="Types do not correspond to columns"):
        sql.Values(target_cols, target_types[:-1], *target_values)

    # Test SELECT INTO with VALUES clause

    select_from = sql.Values(",".join(target_cols), ",".join(target_types), *target_values)
    select_into = sql.SelectInto([sql.column(c) for c in target_cols], into_table_name).select_from(select_from)
    with db_metadata.bind.connect() as conn:
        conn.execute(select_into.execution_options(autocommit=True))

    assert_table(db_metadata, into_table_name, target_cols)
    assert_records(db_metadata, into_table_name, target_data, target_cols)

    into_table = refresh_metadata(db_metadata).tables[into_table_name]

    for idx, col in enumerate(target_cols):
        column_type = str(into_table.columns[col].type).lower()
        target_type = target_types[idx].lower()

        if target_type in ("binary", "text", "unicode"):
            assert column_type == "text"
        elif target_type in ("float",):
            assert column_type == "numeric"
        else:
            assert column_type == target_type

    # Test INSERT INTO with VALUES clause

    insert_from = sql.Values(",".join(target_cols), ",".join(target_types), *target_values)
    insert_vals = sql.Select([sql.column(c) for c in target_cols]).select_from(insert_from)
    insert_into = sql.Insert(into_table).from_select(names=target_cols, select=insert_vals)
    with schema.get_engine().connect() as conn:
        conn.execute(insert_into.execution_options(autocommit=True))

    target_count = len(target_values) * 2
    assert_table(db_metadata, into_table_name, target_cols)
    assert_records(db_metadata, into_table_name, target_data, target_cols, target_count)


def test_table_exists(db_metadata):

    existing = db_metadata.tables

    # Test all tables present in database
    for table_name in existing:
        assert schema.table_exists(existing.get(table_name))
        assert schema.table_exists(table_name)

    assert not schema.table_exists(None)
    assert not schema.table_exists("nope")

    site_table = existing.get(SITE_TABLE_NAME)
    site_table.drop()
    assert not schema.table_exists(site_table)


def test_create_foreign_key(db_metadata):

    site_table = db_metadata.tables[SITE_TABLE_NAME]
    test_table_name = ALL_TEST_TABLES[1]

    # Test with invalid parameters

    with pytest.raises(ValueError, match="No table named"):
        schema.create_foreign_key("nope", "pk", f"{site_table.name}.obj_order")
    with pytest.raises(ValueError, match="Invalid column names"):
        schema.create_foreign_key(site_table.name, "nope", f"{site_table.name}.obj_order")
    with pytest.raises(ValueError, match="No related column named"):
        schema.create_foreign_key(site_table.name, "pk", f"{site_table.name}.nope")

    test_columns = ("pk", "obj_order", "obj_hash", "test_bool")

    # Create an empty test table to link with a foreign key
    test_table = make_table(db_metadata, test_table_name, test_columns)
    assert_table(db_metadata, test_table.name, test_columns)

    # Ensure there are no foreign keys before the operation
    assert len(refresh_metadata(db_metadata).tables[test_table_name].foreign_keys) == 0

    # Test FK creation when a table name is passed in with a column string
    schema.create_foreign_key(test_table_name, "pk", f"{site_table.name}.pk")
    test_table = refresh_metadata(db_metadata).tables[test_table_name]
    assert len(test_table.foreign_keys) == 1
    assert len(test_table.columns.pk.foreign_keys) == 1

    # Test FK creation when a table is passed in with a column object
    schema.create_foreign_key(test_table, "obj_order", site_table.columns.obj_order)
    test_table = refresh_metadata(db_metadata).tables[test_table_name]
    assert len(test_table.foreign_keys) == 2
    assert len(test_table.columns.obj_order.foreign_keys) == 1


def test_drop_foreign_key(db_metadata):

    site_table = db_metadata.tables[SITE_TABLE_NAME]
    test_table_name = ALL_TEST_TABLES[1]

    # Test with invalid parameters

    schema.drop_foreign_key(None, "pk", checkfirst=True)
    with pytest.raises(ValueError, match="No table named"):
        schema.drop_foreign_key(None, "pk")

    schema.drop_foreign_key("nope", "pk", checkfirst=True)
    with pytest.raises(ValueError, match="No table named"):
        schema.drop_foreign_key("nope", "pk")

    schema.drop_foreign_key(site_table.name, "nope", checkfirst=True)
    with pytest.raises(ValueError, match="No such foreign key in table"):
        schema.drop_foreign_key(site_table.name, "nope")

    # Create an empty test table to link with foreign keys

    test_cols = ("pk", "obj_order", "obj_hash")
    test_table = make_table(db_metadata, test_table_name, test_cols)
    assert_table(db_metadata, test_table.name, test_cols)

    fk1 = ForeignKeyConstraint(
        columns=[test_table.columns.pk],
        refcolumns=[site_table.columns.pk],
        name="test_table_first_fkey",
    )
    ddl = AddConstraint(fk1)
    ddl.execute(test_table.bind, ddl.target)

    fk2 = ForeignKeyConstraint(
        columns=[test_table.columns.obj_order],
        refcolumns=[site_table.columns.obj_order],
        name="test_table_next_fkey",
    )
    ddl = AddConstraint(fk2)
    ddl.execute(test_table.bind, ddl.target)

    fk3 = ForeignKeyConstraint(
        columns=[test_table.columns.obj_hash],
        refcolumns=[site_table.columns.obj_hash],
        name="test_table_last_fkey"
    )
    ddl = AddConstraint(fk3)
    ddl.execute(test_table.bind, ddl.target)

    test_table = refresh_metadata(db_metadata).tables[test_table_name]
    assert len(test_table.foreign_keys) == 3
    assert len(test_table.columns.pk.foreign_keys) == 1
    assert len(test_table.columns.obj_order.foreign_keys) == 1
    assert len(test_table.columns.obj_hash.foreign_keys) == 1

    # Test FK deletion when a table name is passed in with a mixed-case FK name
    schema.drop_foreign_key(test_table_name, "test_table_FIRST_fkey")
    test_table = refresh_metadata(db_metadata).tables[test_table_name]
    assert len(test_table.foreign_keys) == 2
    assert len(test_table.columns.pk.foreign_keys) == 0
    assert len(test_table.columns.obj_order.foreign_keys) == 1
    assert len(test_table.columns.obj_hash.foreign_keys) == 1

    # Test FK deletion when a table object is passed in with an FK constraint
    schema.drop_foreign_key(test_table, fk2)
    test_table = refresh_metadata(db_metadata).tables[test_table_name]
    assert len(test_table.foreign_keys) == 1
    assert len(test_table.columns.pk.foreign_keys) == 0
    assert len(test_table.columns.obj_order.foreign_keys) == 0
    assert len(test_table.columns.obj_hash.foreign_keys) == 1

    # Test FK deletion when a table object is passed in with an FK object
    schema.drop_foreign_key(test_table, fk3.elements[0])
    test_table = refresh_metadata(db_metadata).tables[test_table_name]
    assert len(test_table.foreign_keys) == 0
    assert len(test_table.columns.pk.foreign_keys) == 0
    assert len(test_table.columns.obj_order.foreign_keys) == 0
    assert len(test_table.columns.obj_hash.foreign_keys) == 0


def test_alter_column_type(db_metadata):

    table_name = SITE_TABLE_NAME
    inject_sql = "DROP USER 'postgres' IF EXISTS"

    # Test SQL injection code
    with pytest.raises(ValueError, match="Invalid table name"):
        schema.alter_column_type(inject_sql, "test_bool", "int")
    with pytest.raises(ValueError, match="Invalid column name"):
        schema.alter_column_type(table_name, inject_sql, "int")
    with pytest.raises(ValueError, match="Invalid column type"):
        schema.alter_column_type(table_name, "test_bool", inject_sql)
    with pytest.raises(ValueError, match="Invalid column conversion"):
        schema.alter_column_type(table_name, "test_bool", "int", inject_sql)

    # PK column tests

    # Ensure original column is type INTEGER
    original = db_metadata.tables.get(table_name).columns.pk
    assert str(original.type).lower() == "integer"

    # Test change from INTEGER to VARCHAR
    schema.alter_column_type(table_name, "pk", "varchar")
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.pk
    assert str(altered.type).lower() == "varchar"

    # Test change from VARCHAR back to INTEGER
    schema.alter_column_type(table_name, "pk", "int")
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.pk
    assert str(altered.type).lower() == "integer"

    # Bool column tests

    # Ensure original column is type BOOLEAN
    original = db_metadata.tables.get(table_name).columns.test_bool
    assert str(original.type).lower() == "boolean"

    # Test change from BOOLEAN to INTEGER
    schema.alter_column_type(db_metadata.tables[table_name], "test_bool", "int")
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_bool
    assert str(altered.type).lower() == "integer"

    # Test change from INTEGER back to BOOLEAN
    schema.alter_column_type(db_metadata.tables[table_name], "test_bool", "bool")
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_bool
    assert str(altered.type).lower() == "boolean"

    # JSON column tests

    # Ensure original column is type JSONB
    original = db_metadata.tables.get(table_name).columns.test_json
    assert str(original.type).lower() == "jsonb"

    # Test change from JSONB to VARCHAR
    schema.alter_column_type(db_metadata.tables[table_name], "test_json", "varchar")
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_json
    assert str(altered.type).lower() == "varchar"

    # Test change from VARCHAR to BYTEA
    schema.alter_column_type(db_metadata.tables[table_name], "test_json", "bytea")
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_json
    assert str(altered.type).lower() == "bytea"

    # Test change from BYTEA back to JSONB
    schema.alter_column_type(db_metadata.tables[table_name], "test_json", "jsonb")
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_json
    assert str(altered.type).lower() == "jsonb"

    # Test change from JSONB directly to BYTEA
    schema.alter_column_type(db_metadata.tables[table_name], "test_json", "bytea")
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_json
    assert str(altered.type).lower() == "bytea"

    # Test change from BYTEA back to JSONB with USING
    using = "convert_from(test_json,'UTF8')::jsonb"
    schema.alter_column_type(db_metadata.tables[table_name], "test_json", "jsonb", using=using)
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_json
    assert str(altered.type).lower() == "jsonb"

    # Test change from JSONB directly to BYTEA with USING
    using = "test_json::text::bytea"
    schema.alter_column_type(db_metadata.tables[table_name], "test_json", "bytea", using=using)
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_json
    assert str(altered.type).lower() == "bytea"

    # Cross-type JSON conversion tests

    # Test conversion of BOOLEAN to JSONB
    schema.alter_column_type(db_metadata.tables[table_name], "test_bool", "jsonb")
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_bool
    assert str(altered.type).lower() == "jsonb"

    # Test conversion of NUMERIC to JSONB
    schema.alter_column_type(db_metadata.tables[table_name], "obj_order", "jsonb")
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.obj_order
    assert str(altered.type).lower() == "jsonb"

    # Geometry column tests

    # Ensure original column is type GEOMETRY (should be POINT)
    original = db_metadata.tables.get(table_name).columns.test_geom
    assert str(original.type).lower().startswith("geometry")

    # Test change from GEOMETRY to BYTEA
    schema.alter_column_type(db_metadata.tables[table_name], "test_geom", "bytea")
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_geom
    assert str(altered.type).lower() == "bytea"

    # Test change from BYTEA to generic GEOMETRY (without "using")
    schema.alter_column_type(db_metadata.tables[table_name], "test_geom", "geometry")
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_geom
    assert str(altered.type).lower().startswith("geometry")

    # Test change from GEOMETRY to TEXT
    schema.alter_column_type(db_metadata.tables[table_name], "test_geom", "text")
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_geom
    assert str(altered.type).lower() == "text"

    # Test change from TEXT to POINT
    using = "test_geom::geometry(POINT,4326)"
    schema.alter_column_type(db_metadata.tables[table_name], "test_geom", "geometry", using=using)
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_geom
    assert str(altered.type).lower().startswith("geometry")

    # Test change from POINT to LINE
    using = "test_geom::geometry(LINESTRING,4326)"
    schema.alter_column_type(db_metadata.tables[table_name], "test_geom", "geometry", using=using)
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_geom
    assert str(altered.type).lower().startswith("geometry")

    # Test change from LINE to POLYGON
    using = "test_geom::geometry(POLYGON,4326)"
    schema.alter_column_type(db_metadata.tables[table_name], "test_geom", "geometry", using=using)
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_geom
    assert str(altered.type).lower().startswith("geometry")

    # Test change from POLYGON back to BYTEA
    schema.alter_column_type(db_metadata.tables[table_name], "test_geom", "bytea")
    altered = refresh_metadata(db_metadata).tables.get(table_name).columns.test_geom
    assert str(altered.type).lower() == "bytea"


def test_create_column(db_metadata):

    site_table = db_metadata.tables[SITE_TABLE_NAME]
    inject_sql = "DROP USER 'postgres' IF EXISTS"

    # Test SQL injection code
    with pytest.raises(ValueError, match="Invalid table name"):
        schema.create_column(inject_sql, "test_json", "jsonb", default="")
    with pytest.raises(ValueError, match="Invalid column name"):
        schema.create_column(site_table.name, inject_sql, "jsonb", default="")
    with pytest.raises(ValueError, match="Invalid column type"):
        schema.create_column(site_table, "test_json", inject_sql, default="")
    with pytest.raises(ValueError, match="Invalid default value"):
        schema.create_column(site_table.name, "test_json", "jsonb", default=inject_sql)

    # Test when column already exists

    column = "test_bool"
    assert column in site_table.columns

    schema.create_column(site_table.name, "test_bool", "bool", default=False, checkfirst=True)
    with pytest.raises(Exception):
        schema.create_column(site_table.name, "test_bool", "bool", default=False, checkfirst=False)

    # Test column creation

    column = "test_text"
    schema.create_column(site_table.name, column, "varchar", default="nothing")
    newcol = refresh_metadata(db_metadata).tables[site_table.name].columns.get(column)
    assert column == newcol.name
    assert str(newcol.type).lower().startswith("varchar")
    assert not newcol.nullable

    column = "test_true"
    schema.create_column(site_table.name, column, "bool", default=True)
    newcol = refresh_metadata(db_metadata).tables[site_table.name].columns.get(column)
    assert column == newcol.name
    assert str(newcol.type).lower().startswith("bool")
    assert not newcol.nullable

    column = "test_int"
    schema.create_column(site_table.name, column, sqltypes.Integer, default=50)
    newcol = refresh_metadata(db_metadata).tables[site_table.name].columns.get(column)
    assert column == newcol.name
    assert str(newcol.type).lower().startswith("integer")
    assert not newcol.nullable

    column = "test_num"
    schema.create_column(site_table.name, column, "decimal", default=3.14159)
    newcol = refresh_metadata(db_metadata).tables[site_table.name].columns.get(column)
    assert column == newcol.name
    assert str(newcol.type).lower().startswith("numeric")
    assert not newcol.nullable

    column = "test_jsonb"
    schema.create_column(site_table.name, column, postgresql.json.JSONB, nullable=True)
    newcol = refresh_metadata(db_metadata).tables[site_table.name].columns.get(column)
    assert column == newcol.name
    assert str(newcol.type).lower().startswith("json")
    assert newcol.nullable

    column = "test_point"
    schema.create_column(site_table.name, column, "geometry(POINT,4326)", nullable=True)
    newcol = refresh_metadata(db_metadata).tables[site_table.name].columns.get(column)
    assert column == newcol.name
    assert str(newcol.type).lower().startswith("geometry")
    assert newcol.nullable


def test_drop_column(db_metadata):

    site_table = db_metadata.tables[SITE_TABLE_NAME]
    inject_sql = "DROP USER 'postgres' IF EXISTS"

    # Test SQL injection code
    with pytest.raises(ValueError, match="Invalid table name"):
        schema.drop_column(inject_sql, "test_bool")
    with pytest.raises(ValueError, match="Invalid column name"):
        schema.drop_column(site_table.name, inject_sql)

    # Test when the column doesn't exist

    column = "nope"
    assert column not in site_table.columns
    schema.drop_column(site_table.name, column, checkfirst=True)
    assert column not in refresh_metadata(db_metadata).tables[site_table.name].columns

    # Test for the presence and removal of a simple column

    column = "test_bool"
    assert column in site_table.columns
    schema.drop_column(site_table.name, column)
    assert column not in refresh_metadata(db_metadata).tables[site_table.name].columns

    # Test for the presence and removal of an indexed column

    column = "obj_order"
    assert column in site_table.columns
    schema.drop_column(site_table, column)
    assert column not in refresh_metadata(db_metadata).tables[site_table.name].columns


def test_create_tsvector_column(db_metadata):

    site_table = db_metadata.tables[SITE_TABLE_NAME]
    inject_sql = "DROP USER 'postgres' IF EXISTS"

    # At least test SQL injection code
    with pytest.raises(ValueError, match="Invalid table name"):
        schema.create_tsvector_column(inject_sql, "searchable", ["obj_hash"])
    with pytest.raises(ValueError, match="Invalid column name"):
        schema.create_tsvector_column(site_table, inject_sql, ["obj_hash"], "tsvector_index")
    with pytest.raises(ValueError, match="Invalid column names"):
        schema.create_tsvector_column(site_table.name, "searchable", ["obj_hash", inject_sql], "tsvector_index")

    # Test when column already exists

    with pytest.raises(Exception):
        schema.create_tsvector_column(site_table.name, "site_addr", SEARCHABLE_COLS)

    # Test column creation on all searchable columns

    col_name = "generated_search_col"
    idx_name = "generated_tsvector_index"
    schema.create_tsvector_column(site_table.name, col_name, SEARCHABLE_COLS, index_name=idx_name)

    newcol = refresh_metadata(db_metadata).tables[site_table.name].columns.get(col_name)
    assert col_name == newcol.name
    assert str(newcol.type).lower().startswith("tsvector")

    # TSVECTOR indexes are not loaded from database via table reflection
    assert_index(db_metadata, site_table.name, index_name=idx_name)

    # Test column creation on a subset of searchable columns

    col_name = "generated_site_col"
    idx_name = "generated_site_index"
    site_cols = ",".join(col for col in SEARCHABLE_COLS if col.startswith("site_"))
    schema.create_tsvector_column(site_table.name, col_name, site_cols, index_name=idx_name)

    newcol = refresh_metadata(db_metadata).tables[site_table.name].columns.get(col_name)
    assert col_name == newcol.name
    assert str(newcol.type).lower().startswith("tsvector")

    # TSVECTOR indexes are not loaded from database via table reflection
    assert_index(db_metadata, site_table.name, index_name=idx_name)


def test_query_json_keys(db_metadata):

    site_table = db_metadata.tables[SITE_TABLE_NAME]

    # Test with invalid parameters

    with pytest.raises(ValueError, match="No table named"):
        sql.query_json_keys("nope", "site_addr", {"state": "NY"})
    with pytest.raises(ValueError, match="Invalid column name"):
        sql.query_json_keys(SITE_TABLE_NAME, "nope", {"state": "NY"})

    # Test a query that will return all columns for all records
    limit = len(SITE_TEST_DATA)
    search_results = sql.query_json_keys(site_table, "site_json", {"city": "ALBANY", "state": "NY"})
    assert limit == len({result["pk"] for result in search_results})
    assert all(len(result) == len(SITE_TABLE_COLS) for result in search_results)

    # Test limiting a query that will return all columns for all records
    limit = round(len(SITE_TEST_DATA) / 2)
    search_results = sql.query_json_keys(site_table, "site_json", {"city": "ALBANY", "state": "NY"}, limit=limit)
    assert limit == len({result["pk"] for result in search_results})
    assert all(len(result) == len(SITE_TABLE_COLS) for result in search_results)

    # Test an excessive limit on a query that will return all columns for all records
    limit = len(SITE_TEST_DATA)
    search_results = sql.query_json_keys(site_table, "site_json", {"city": "ALBANY", "state": "NY"}, limit=(limit * 2))
    assert limit == len({result["pk"] for result in search_results})
    assert all(len(result) == len(SITE_TABLE_COLS) for result in search_results)

    # Test a single matching address as string
    query, matches = json.dumps({"number": "7550", "street": "STATE ST STE CML5"}), 1
    search_results = sql.query_json_keys(site_table.name, "site_json", query)
    assert matches == len({result["pk"] for result in search_results})

    # Test an invalid address
    query, matches = {"street": "nope"}, 0
    search_results = sql.query_json_keys(site_table.name, "site_json", query)
    assert matches == len({result["pk"] for result in search_results})

    # Test an empty address
    query, matches = '""', 0
    search_results = sql.query_json_keys(site_table.name, "site_json", query)
    assert matches == len({result["pk"] for result in search_results})


def test_query_tsvector_columns(db_metadata):

    site_table = db_metadata.tables[SITE_TABLE_NAME]

    # Test with invalid parameters

    with pytest.raises(ValueError, match="No table named"):
        sql.query_tsvector_columns("nope", "site_addr", "NY")
    with pytest.raises(ValueError, match="Invalid column names"):
        sql.query_tsvector_columns(site_table, "nope", "NY")

    # Test queries that match all records
    limit = len(SITE_TEST_DATA)
    for query in ("ALBANY", "NY", "STATE", "12224"):
        assert limit == len(sql.query_tsvector_columns(SITE_TABLE_NAME, SEARCHABLE_COLS, query))

    # Test that limit applies to queries matching all records
    for limit, query in enumerate(("ALBANY", "NY", "STATE", "12224")):
        search_results = sql.query_tsvector_columns(site_table, SEARCHABLE_COLS, query, limit=limit)
        assert limit == len({result["pk"] for result in search_results})

    # Test an excessive limit on a query that matches all records
    query, matches = "NY", len(SITE_TEST_DATA)
    search_results = sql.query_tsvector_columns(site_table.name, SEARCHABLE_COLS, query, limit=(matches * 2))
    assert matches == len({result["pk"] for result in search_results})
    assert all(len(result) == len(SITE_TABLE_COLS) for result in search_results)

    # Test queries that match a single record

    query, matches = "0-0-7350-12224", 1
    search_results = sql.query_tsvector_columns(site_table.name, SEARCHABLE_COLS, query)
    assert matches == len({result["pk"] for result in search_results})
    assert all(len(result) == len(SITE_TABLE_COLS) for result in search_results)

    query, matches = "7550 STATE ST STE CML5", 1
    search_results = sql.query_tsvector_columns(site_table.name, SEARCHABLE_COLS, query)
    assert matches == len({result["pk"] for result in search_results})
    assert all(len(result) == len(SITE_TABLE_COLS) for result in search_results)

    terms = "7550 STATE ST STE CML5".split()[::-1]
    query, matches = " ".join(terms), 1
    search_results = sql.query_tsvector_columns(site_table.name, SEARCHABLE_COLS, query)
    assert matches == len({result["pk"] for result in search_results})
    assert all(len(result) == len(SITE_TABLE_COLS) for result in search_results)

    query, matches = "8-4-7550-12224 7550 STATE ST STE CML5 ALBANY NY 12224", 1
    search_results = sql.query_tsvector_columns(site_table.name, SEARCHABLE_COLS, query)
    assert matches == len({result["pk"] for result in search_results})
    assert all(len(result) == len(SITE_TABLE_COLS) for result in search_results)

    terms = "8-4-7550-12224 7550 STATE ST STE CML5 ALBANY NY 12224".split()[::-1]
    query, matches = " ".join(terms), 1
    search_results = sql.query_tsvector_columns(site_table.name, SEARCHABLE_COLS, query)
    assert matches == len({result["pk"] for result in search_results})
    assert all(len(result) == len(SITE_TABLE_COLS) for result in search_results)

    # Test queries that don't match any records

    query, matches = "nope", 0
    search_results = sql.query_tsvector_columns(site_table.name, SEARCHABLE_COLS, query)
    assert matches == len({result["pk"] for result in search_results})

    query, matches = "STATE ST", 0
    search_results = sql.query_tsvector_columns(site_table.name, "obj_hash", query)
    assert matches == len({result["pk"] for result in search_results})

    query, matches = "0-0-7350-12224", 0
    search_results = sql.query_tsvector_columns(site_table.name, "site_addr", query)
    assert matches == len({result["pk"] for result in search_results})



import subprocess

from setuptools import Command, setup


class RunTests(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        errno = subprocess.call(["py.test", "pg_database/tests/tests.py", "--cov=pg_database", "--cov-branch"])
        raise SystemExit(errno)


with open("README.md") as readme:
    long_description = readme.read()

setup(
    url="https://github.com/consbio/pg-database-utils",
    name="pg-database-utils",
    description="A suite of utilities for PostgreSQL database queries and operations built on sqlalchemy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="postgres,postgresql,utils,utilities,pg_database,pg_database_utils,sqlalchemy",
    version="0.8.0",
    license="BSD",
    packages=[
        "pg_database", "pg_database.tests"
    ],
    install_requires=[
        "frozendict>=1.2", "psycopg2-binary>=2.7.7", "sqlalchemy>=1.3.0", "GeoAlchemy2>=0.7.0"
    ],
    tests_require=["pytest", "pytest-cov"],
    cmdclass={"test": RunTests}
)
